from typing import Optional

import torch

from ..spectral import spectral_density
from ..types import FrequencySeries1to3d, TimeSeries1to3d


class FittableTransform(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.built = False

    def build(self, **params):
        state_dict = self.state_dict()
        for name, value in params.items():
            state_dict[name].copy_(value)
        self.built = True

    def _check_built(self):
        if not self.built:
            raise ValueError(
                "Must fit parameters of {} transform to data "
                "before calling forward step".format(self.__class__.__name__)
            )

    def __call__(self, *args, **kwargs):
        self._check_built()
        return super().__call__(*args, **kwargs)

    def _load_from_state_dict(self, *args):
        # keep track of number of error messages to see
        # if trying to load these weights causes another
        num_errs = len(args[-1])
        ret = super()._load_from_state_dict(*args)

        # if we didn't create any new errors, then
        # assume that we built correctly
        if len(args[-1]) == num_errs:
            self.built = True
        return ret


class FittableSpectralTransform(FittableTransform):
    def normalize_psd(
        self,
        x: TimeSeries1to3d,
        sample_rate: float,
        num_freqs: int,
        fftlength: Optional[float] = None,
        overlap: Optional[float] = None,
    ) -> FrequencySeries1to3d:
        # if we specified an FFT length, convert
        # the (assumed) time-domain data to the
        # frequency domain
        if fftlength is not None:
            nperseg = int(fftlength * sample_rate)

            overlap = overlap or fftlength / 2
            nstride = nperseg - int(overlap * sample_rate)

            window = torch.hann_window(nperseg, dtype=torch.float64)
            scale = 1.0 / (sample_rate * (window**2).sum())
            x = spectral_density(
                x,
                nperseg=nperseg,
                nstride=nstride,
                window=window,
                scale=scale,
            )

        # add two dummy dimensions in case we need to interpolate
        # the frequency dimension, since `interpolate` expects
        # a (batch, channel, spatial) formatted tensor as input
        x = x.view(1, 1, -1)
        if x.size(-1) != num_freqs:
            x = torch.nn.functional.interpolate(x, size=(num_freqs,))
        return x[0, 0]



"""
Module containing callables classes for generating samples
from specified distributions. Each callable should map from
an integer `N` to a 1D torch `Tensor` containing `N` samples
from the corresponding distribution.
"""
import math
from typing import Optional

import torch
import torch.distributions as dist
from jaxtyping import Float
from torch import Tensor


class Cosine(dist.Distribution):
    """
    Cosine distribution based on
    ``torch.distributions.TransformedDistribution``.
    """

    arg_constraints = {}

    def __init__(
        self,
        low: float = -math.pi / 2,
        high: float = math.pi / 2,
        validate_args=None,
    ):
        batch_shape = torch.Size()
        super().__init__(batch_shape, validate_args=validate_args)
        self.low = torch.as_tensor(low)
        self.high = torch.as_tensor(high)
        self.norm = 1 / (torch.sin(self.high) - torch.sin(self.low))

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> Tensor:
        u = torch.rand(sample_shape, device=self.low.device)
        return torch.arcsin(u / self.norm + torch.sin(self.low))

    def log_prob(self, value: float) -> Float[Tensor, ""]:
        value = torch.as_tensor(value)
        inside_range = (value >= self.low) & (value <= self.high)
        return value.cos().log() * inside_range


class Sine(dist.TransformedDistribution):
    """
    Sine distribution based on
    ``torch.distributions.TransformedDistribution``.
    """

    def __init__(
        self,
        low: float = 0.0,
        high: float = math.pi,
        validate_args=None,
    ):
        low = torch.as_tensor(low)
        high = torch.as_tensor(high)
        base_dist = Cosine(
            low - torch.pi / 2, high - torch.pi / 2, validate_args
        )

        super().__init__(
            base_dist,
            [
                dist.AffineTransform(
                    loc=torch.pi / 2,
                    scale=1,
                )
            ],
            validate_args=validate_args,
        )


class LogUniform(dist.TransformedDistribution):
    """
    Sample from a log uniform distribution
    """

    def __init__(self, low: float, high: float, validate_args=None):
        base_dist = dist.Uniform(
            torch.as_tensor(low).log(),
            torch.as_tensor(high).log(),
            validate_args,
        )
        super().__init__(
            base_dist,
            [dist.ExpTransform()],
            validate_args=validate_args,
        )


class LogNormal(dist.LogNormal):
    def __init__(
        self,
        mean: float,
        std: float,
        low: Optional[float] = None,
        validate_args=None,
    ):
        self.low = low
        super().__init__(loc=mean, scale=std, validate_args=validate_args)

    def support(self):
        if self.low is not None:
            return dist.constraints.greater_than(self.low)


class PowerLaw(dist.TransformedDistribution):
    """
    Sample from a power law distribution,
    .. math::
        p(x) \approx x^{\alpha}.

    Index alpha cannot be 0, since it is equivalent to a Uniform distribution.
    This could be used, for example, as a universal distribution of
    signal-to-noise ratios (SNRs) from uniformly volume distributed
    sources
    .. math::

       p(\rho) = 3*\rho_0^3 / \rho^4

    where :math:`\rho_0` is a representative minimum SNR
    considered for detection. See, for example,
    `Schutz (2011) <https://arxiv.org/abs/1102.5421>`_.
    Or, for example, ``index=2`` for uniform in Euclidean volume.
    """

    support = dist.constraints.nonnegative

    def __init__(
        self, minimum: float, maximum: float, index: int, validate_args=None
    ):
        if index == 0:
            raise RuntimeError("Index of 0 is the same as Uniform")
        elif index == -1:
            base_min = torch.as_tensor(minimum).log()
            base_max = torch.as_tensor(maximum).log()
            transforms = [dist.ExpTransform()]
        else:
            index_plus = index + 1
            base_min = minimum**index_plus / index_plus
            base_max = maximum**index_plus / index_plus
            transforms = [
                dist.AffineTransform(loc=0, scale=index_plus),
                dist.PowerTransform(1 / index_plus),
            ]
        base_dist = dist.Uniform(base_min, base_max, validate_args=False)
        super().__init__(
            base_dist,
            transforms,
            validate_args=validate_args,
        )


class DeltaFunction(dist.Distribution):
    arg_constraints = {}

    def __init__(
        self,
        peak: float = 0.0,
        validate_args=None,
    ):
        batch_shape = torch.Size()
        super().__init__(batch_shape, validate_args=validate_args)
        self.peak = torch.as_tensor(peak)

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> Tensor:
        return self.peak * torch.ones(
            sample_shape, device=self.peak.device, dtype=torch.float32
        )



from collections.abc import Callable, Sequence
from typing import Optional

import torch
from torch import Tensor

from .base import Autoencoder
from .skip_connection import SkipConnection
from .utils import match_size

Module = Callable[[...], torch.nn.Module]


class ConvBlock(Autoencoder):
    def __init__(
        self,
        in_channels: int,
        encode_channels: int,
        kernel_size: int,
        stride: int = 1,
        groups: int = 1,
        activation: torch.nn.Module = torch.nn.ReLU(),
        norm: Module = torch.nn.BatchNorm1d,
        decode_channels: Optional[int] = None,
        output_activation: Optional[torch.nn.Module] = None,
        skip_connection: Optional[SkipConnection] = None,
    ) -> None:
        super().__init__(skip_connection=None)

        self.kernel_size = kernel_size
        self.padding = int((kernel_size - 1) // 2)
        self.stride = stride

        out_channels = encode_channels * groups
        self.encode_layer = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.padding,
            bias=False,
            groups=groups,
        )

        decode_channels = decode_channels or in_channels
        in_channels = encode_channels * groups
        if skip_connection is not None:
            in_channels = skip_connection.get_out_channels(in_channels)
        self.decode_layer = torch.nn.ConvTranspose1d(
            in_channels,
            decode_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.padding,
            bias=False,
            groups=groups,
        )

        self.activation = activation
        if output_activation is not None:
            self.output_activation = output_activation
        else:
            self.output_activation = activation

        self.encode_norm = norm(out_channels)
        self.decode_norm = norm(decode_channels)

    def encode(self, X: Tensor) -> Tensor:
        X = self.encode_layer(X)
        X = self.encode_norm(X)
        return self.activation(X)

    def decode(self, X: Tensor) -> Tensor:
        X = self.decode_layer(X)
        X = self.decode_norm(X)
        return self.output_activation(X)


class ConvolutionalAutoencoder(Autoencoder):
    """
    Build a stack of convolutional autoencoder layer
    blocks. The output of each decoder layer will
    match the shape of the input to its corresponding
    encoder layer, except for the last decoder which
    can have an arbitrary number of channels specified
    by `decode_channels`.

    All layers also share the same `activation` except
    for the last decoder layer, which can have an
    arbitrary `output_activation`.
    """

    def __init__(
        self,
        in_channels: int,
        encode_channels: Sequence[int],
        kernel_size: int,
        stride: int = 1,
        groups: int = 1,
        activation: torch.nn.Module = torch.nn.ReLU(),
        output_activation: Optional[torch.nn.Module] = None,
        norm: Module = torch.nn.BatchNorm1d,
        decode_channels: Optional[int] = None,
        skip_connection: Optional[SkipConnection] = None,
    ) -> None:
        # TODO: how to do this dynamically? Maybe the base
        # architecture looks for overlapping arguments between
        # this and the skip connection class and then provides them?
        # if skip_connection is not None:
        #     skip_connection = skip_connection(groups)
        super().__init__(skip_connection=skip_connection)

        output_activation = output_activation or activation
        for i, channels in enumerate(encode_channels):
            # All intermediate layers should decode to
            # the same number of channels. The last decoder
            # should decode to whatever number of channels
            # was specified, even if it's `None` (in which
            # case it will just be in_channels anyway)
            decode = in_channels if i else decode_channels

            # don't have the middle layer skip to itself
            # TODO: wait I don't think this makes sense.
            # j = len(encode_channels) - 1 - i
            # connect = skip_connection if j else None
            connect = skip_connection

            # all intermediate layers should use the same
            # activation. Only the last decoder should have
            # a potentially different activation
            out_act = None if i else output_activation

            block = ConvBlock(
                in_channels,
                channels,
                kernel_size,
                stride,
                groups,
                activation=activation,
                norm=norm,
                decode_channels=decode,
                skip_connection=connect,
                output_activation=out_act,
            )
            self.blocks.append(block)
            in_channels = channels * groups

    def decode(
        self, *X, states=None, input_size: Optional[int] = None
    ) -> Tensor:
        X = super().decode(*X, states=states)
        if input_size is not None:
            return match_size(X, input_size)
        return X

    def forward(self, X: Tensor) -> Tensor:
        input_size = X.size(-1)
        X = super().forward(X)
        return match_size(X, input_size)



"""
In large part lifted from
https://github.com/pytorch/vision/blob/main/torchvision/models/resnet.py
but with 1d convolutions and arbitrary kernel sizes, and a
default norm layer that makes more sense for most GW applications
where training-time statistics are entirely arbitrary due to
simulations.
"""

from typing import Callable, List, Literal, Optional

import torch
import torch.nn as nn
from torch import Tensor

from ..norm import GroupNorm1DGetter, NormLayer


def convN(
    in_planes: int,
    out_planes: int,
    kernel_size: int = 3,
    stride: int = 1,
    groups: int = 1,
    dilation: int = 1,
) -> nn.Conv1d:
    """1d convolution with padding"""
    if not kernel_size % 2:
        raise ValueError("Can't use even sized kernels")

    return nn.Conv1d(
        in_planes,
        out_planes,
        kernel_size=kernel_size,
        stride=stride,
        padding=dilation * int(kernel_size // 2),
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv1d:
    """Kernel-size 1 convolution"""
    return nn.Conv1d(
        in_planes, out_planes, kernel_size=1, stride=stride, bias=False
    )


class BasicBlock(nn.Module):
    """Defines the structure of the blocks used to build the ResNet"""

    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        kernel_size: int = 3,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:

        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm1d
        if groups != 1 or base_width != 64:
            raise ValueError(
                "BasicBlock only supports groups=1 and base_width=64"
            )
        if dilation > 1:
            raise NotImplementedError(
                "Dilation > 1 not supported in BasicBlock"
            )

        # Both self.conv1 and self.downsample layers
        # downsample the input when stride != 1
        self.conv1 = convN(inplanes, planes, kernel_size, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = convN(planes, planes, kernel_size)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    """
    Bottleneck blocks implement one extra convolution
    compared to basic blocks. In this layers, the `planes`
    parameter is generally meant to _downsize_ the number
    of feature maps first, which then get expanded out to
    `planes * Bottleneck.expansion` feature maps at the
    output of the layer.
    """

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        kernel_size: int = 3,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[NormLayer] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm1d

        width = int(planes * (base_width / 64.0)) * groups

        # conv1 does no downsampling, just reduces the number of
        # feature maps from inplanes to width (where width == planes)
        # if groups == 1 and base_width == 64
        self.conv1 = convN(inplanes, width, kernel_size)
        self.bn1 = norm_layer(width)

        # conv2 keeps the same number of feature maps,
        # but downsamples along the time axis if stride
        # or dilation > 1
        self.conv2 = convN(width, width, kernel_size, stride, groups, dilation)
        self.bn2 = norm_layer(width)

        # conv3 expands the feature maps back out to planes * expansion
        self.conv3 = conv1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet1D(nn.Module):
    """1D ResNet architecture

    Simple extension of ResNet to 1D convolutions with
    arbitrary kernel sizes to support the longer timeseries
    used in BBH detection.

    Args:
        in_channels:
            The number of channels in input tensor.
        layers:
            A list representing the number of residual
            blocks to include in each "layer" of the
            network. Total layers (e.g. 50 in ResNet50)
            is `2 + sum(layers) * factor`, where factor
            is `2` for vanilla `ResNet` and `3` for
            `BottleneckResNet`.
        kernel_size:
            The size of the convolutional kernel to
            use in all residual layers. _NOT_ the size
            of the input kernel to the network, which
            is determined at run-time.
        zero_init_residual:
            Flag indicating whether to initialize the
            weights of the batch-norm layer in each block
            to 0 so that residuals are initialized as
            identities. Can improve training results.
        groups:
            Number of convolutional groups to use in all
            layers. Grouped convolutions induce local
            connections between feature maps at subsequent
            layers rather than global. Generally won't
            need this to be >1, and wil raise an error if
            >1 when using vanilla `ResNet`.
        width_per_group:
            Base width of each of the feature map groups,
            which is scaled up by the typical expansion
            factor at each layer of the network. Meaningless
            for vanilla `ResNet`.
        stride_type:
            Whether to achieve downsampling on the time axis
            by strided or dilated convolutions for each layer.
            If left as `None`, strided convolutions will be
            used at each layer. Otherwise, `stride_type` should
            be one element shorter than `layers` and indicate either
            `stride` or `dilation` for each layer after the first.
    """

    block = BasicBlock

    def __init__(
        self,
        in_channels: int,
        layers: List[int],
        classes: int,
        kernel_size: int = 3,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        stride_type: Optional[List[Literal["stride", "dilation"]]] = None,
        norm_layer: Optional[NormLayer] = None,
    ) -> None:
        super().__init__()

        self.inplanes = 64
        self.dilation = 1

        # default to using InstanceNorm if no
        # norm layer is provided explicitly
        self._norm_layer = norm_layer or GroupNorm1DGetter()

        # TODO: should we support passing a single string
        # for simplicity here?
        if stride_type is None:
            # each element in the tuple indicates if we should replace
            # the stride with a dilated convolution instead
            stride_type = ["stride"] * (len(layers) - 1)
        if len(stride_type) != (len(layers) - 1):
            raise ValueError(
                "'stride_type' should be None or a "
                "{}-element tuple, got {}".format(len(layers) - 1, stride_type)
            )

        self.groups = groups
        self.base_width = width_per_group

        # start with a basic conv-bn-relu-maxpool block
        # to reduce the dimensionality before the heavy
        # lifting starts
        self.conv1 = nn.Conv1d(
            in_channels,
            self.inplanes,
            kernel_size=7,
            stride=2,
            padding=3,
            bias=False,
        )
        self.bn1 = self._norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool1d(kernel_size=3, stride=2, padding=1)

        # now create layers of residual blocks where each
        # layer uses the same number of feature maps for
        # all its blocks (some power of 2 times 64).
        # Don't downsample along the time axis in the first
        # layer, but downsample in all the rest (either by
        # striding or dilating depending on the stride_type
        # argument)
        residual_layers = [self._make_layer(64, layers[0], kernel_size)]
        it = zip(layers[1:], stride_type)
        for i, (num_blocks, stride) in enumerate(it):
            block_size = 64 * 2 ** (i + 1)
            layer = self._make_layer(
                block_size,
                num_blocks,
                kernel_size,
                stride=2,
                stride_type=stride,
            )
            residual_layers.append(layer)
        self.residual_layers = nn.ModuleList(residual_layers)

        # Average pool over each feature map to create a
        # single value for each feature map that we'll use
        # in the fully connected head
        self.avgpool = nn.AdaptiveAvgPool1d(1)

        # use a fully connected layer to map from the
        # feature maps to the binary output that we need
        self.fc = nn.Linear(block_size * self.block.expansion, classes)

        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu"
                )
            elif isinstance(m, (nn.BatchNorm1d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros,
        # and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to
        # https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(
        self,
        planes: int,
        blocks: int,
        kernel_size: int = 3,
        stride: int = 1,
        stride_type: Literal["stride", "dilation"] = "stride",
    ) -> nn.Sequential:
        block = self.block
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation

        if stride_type == "dilation":
            self.dilation *= stride
            stride = 1
        elif stride_type != "stride":
            raise ValueError("Unknown stride type {stride}")

        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                kernel_size,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    kernel_size,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        for layer in self.residual_layers:
            x = layer(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


# TODO: implement as arg of ResNet instead?
class BottleneckResNet1D(ResNet1D):
    """A version of ResNet that uses bottleneck blocks"""

    block = Bottleneck



import torch
from jaxtyping import Float
from torch import Tensor

from ..types import TimeSeries1to3d
from ..utils.slicing import unfold_windows


class ShiftedPearsonCorrelation(torch.nn.Module):
    """
    Compute the [Pearson correlation]
    (https://en.wikipedia.org/wiki/Pearson_correlation_coefficient)
    for two equal-length timeseries over a pre-defined number of time
    shifts in each direction. Useful for when you want a
    correlation, but not over every possible shift (i.e.
    a convolution).

    The number of dimensions of the second timeseries `y`
    passed at call time should always be less than or equal
    to the number of dimensions of the first timeseries `x`,
    and each dimension should match the corresponding one of
    `x` in  reverse order (i.e. if `x` has shape `(B, C, T)`
    then `y` should either have shape `(T,)`, `(C, T)`, or
    `(B, C, T)`).

    Note that no windowing to either timeseries is applied
    at call time. Users should do any requisite windowing
    beforehand.

    TODOs:
    - Should we perform windowing?
    - Should we support stride > 1?

    Args:
        max_shift:
            The maximum number of 1-step time shifts in
            each direction over which to compute the
            Pearson coefficient. Output shape will then
            be `(2 * max_shifts + 1, B, C)`.
    """

    def __init__(self, max_shift: int) -> None:
        super().__init__()
        self.max_shift = max_shift

    def _shape_checks(self, x: TimeSeries1to3d, y: TimeSeries1to3d):
        if x.ndim > 3:
            raise ValueError(
                "Tensor x can only have up to 3 dimensions "
                f"to compute ShiftedPearsonCorrelation. Found {x.ndim}."
            )
        elif y.ndim > x.ndim:
            raise ValueError(
                "y may not have more dimensions that x for "
                "ShiftedPearsonCorrelation, but found shapes "
                "{} and {}".format(y.shape, x.shape)
            )
        for dim in range(y.ndim):
            if y.size(-dim - 1) != x.size(-dim - 1):
                raise ValueError(
                    "x and y expected to have same size along "
                    "last dimensions, but found shapes {} and {}".format(
                        x.shape, y.shape
                    )
                )

    def forward(
        self, x: TimeSeries1to3d, y: TimeSeries1to3d
    ) -> Float[Tensor, "windows ..."]:
        self._shape_checks(x, y)
        dim = x.size(-1)

        # pad x along time dimension so that it has shape
        # batch x channels x (time + 2 * max_shift)
        pad = (self.max_shift, self.max_shift)
        x = torch.nn.functional.pad(x, pad)

        # num_windows x batch x channels x time
        x = unfold_windows(x, dim, 1)

        # now compute the correlation between each window
        # of x and the single window of y. Start by de-meaning
        x = x - x.mean(-1, keepdims=True)
        y = y - y.mean(-1, keepdims=True)

        # apply formula and sum along time dimension to give final shape
        # num_windows x batch x channels
        corr = (x * y).sum(axis=-1)
        norm = (x**2).sum(-1) * (y**2).sum(-1)

        return corr / norm**0.5



from typing import Optional

import torch
from jaxtyping import Float
from torch import Tensor

from .transform import FittableTransform


class ChannelWiseScaler(FittableTransform):
    """Scale timeseries channels to be zero mean unit variance

    Scales timeseries channels by the mean and standard
    deviation of the channels of the timeseries used to
    fit the module. To reverse the scaling, provide the
    `reverse=True` keyword argument at call time.
    By default, the scaling parameters are set to zero mean
    and unit variance, amounting to an identity transform.

    Args:
        num_channels:
            The number of channels of the target timeseries.
            If left as `None`, the timeseries will be assumed
            to be 1D (single channel).
    """

    def __init__(self, num_channels: Optional[int] = None) -> None:
        super().__init__()

        shape = (num_channels or 1,)
        if num_channels is not None:
            shape += (1,)

        mean = torch.zeros(shape)
        std = torch.ones(shape)
        self.register_buffer("mean", mean)
        self.register_buffer("std", std)

    def fit(
        self, X: Float[Tensor, "... time"], std_reg: Optional[float] = 0.0
    ) -> None:
        """Fit the scaling parameters to a timeseries

        Computes the channel-wise mean and standard deviation
        of the timeseries `X` and sets these values to the
        `mean` and `std` parameters of the scaler.
        """

        if X.ndim == 1:
            assert self.mean.ndim == self.std.ndim == 1
            mean = X.mean(dim=0, keepdim=True)
            # default for torch is to include bessel correction
            std = X.std(dim=0, correction=0, keepdim=True)
        elif X.ndim == 2:
            assert self.mean.ndim == self.std.ndim == 2
            assert len(X) == self.mean.size(0) == self.std.size(0)
            mean = X.mean(dim=-1, keepdim=True)
            std = X.std(dim=-1, correction=0, keepdim=True)
        else:
            raise ValueError(
                "Can't fit channel wise mean and standard deviation "
                "from tensor of shape {}".format(X.shape)
            )
        std += std_reg * torch.ones_like(std)
        super().build(mean=mean, std=std)

    def forward(
        self, X: Float[Tensor, "... time"], reverse: bool = False
    ) -> Float[Tensor, "... time"]:
        if not reverse:
            return (X - self.mean) / self.std
        else:
            return self.std * X + self.mean



import torch
from torch import Tensor


def match_size(X: Tensor, target_size: int) -> Tensor:
    diff = target_size - X.size(-1)
    left = int(diff // 2)
    right = diff - left

    if diff > 0:
        return torch.nn.functional.pad(X, (left, right))
    elif diff < 0:
        right = -right or None
        return X[:, :, -left:right]
    return X



import numpy as np
import torch

from ml4gw.constants import PI, C, G, m_per_Mpc
from ml4gw.types import BatchTensor


class Ringdown(torch.nn.Module):
    """
    Callable class for generating ringdown waveforms.

    Args:
        sample_rate: Sample rate of waveform
        duration: Duration of waveform
    """

    def __init__(self, sample_rate: float, duration: float):
        super().__init__()
        # determine times based on requested duration and sample rate
        # and shift so that the waveform is centered at t=0

        num = int(duration * sample_rate)
        times = torch.arange(num, dtype=torch.float64) / sample_rate

        self.register_buffer("times", times)

    def forward(
        self,
        frequency: BatchTensor,
        quality: BatchTensor,
        epsilon: BatchTensor,
        phase: BatchTensor,
        inclination: BatchTensor,
        distance: BatchTensor,
    ):
        """
        Generate ringdown waveform based on the damped sinusoid equation.

        Args:
            frequency:
                Central frequency of the ringdown waveform in Hz
            quality:
                Quality factor of the ringdown waveform
            epsilon:
                Fraction of black hole's mass radiated as gravitational waves
            phase:
                Initial phase of the ringdown waveform in rad
            inclination:
                Inclination angle of the source in rad
            distance:
                Distance to the source in Mpc
        Returns:
            Tensors of cross and plus polarizations
        """

        # add dimension for calculating waveforms in batch
        frequency = frequency.view(-1, 1)
        quality = quality.view(-1, 1)
        epsilon = epsilon.view(-1, 1)
        phase = phase.view(-1, 1)
        inclination = inclination.view(-1, 1)
        distance = distance.view(-1, 1)

        # convert Mpc to m
        distance = distance * m_per_Mpc

        # ensure all inputs are on the same device
        pi = torch.tensor([PI], device=frequency.device)

        # Calculate spin and mass
        spin = 1 - (2 / quality) ** (20 / 9)
        mass = (
            (1 / (2 * pi))
            * (C**3 / (G * frequency))
            * (1 - 0.63 * (2 / quality) ** (2 / 3))
        )

        # Calculate amplitude
        F_Q = 1 + ((7 / 24) / quality**2)
        g_a = 1 - 0.63 * (1 - spin) ** (3 / 10)
        amplitude = (
            np.sqrt(5 / 2)
            * epsilon
            * (G * mass / (C) ** 2)
            * quality ** (-0.5)
            * F_Q ** (-0.5)
            * g_a ** (-0.5)
        )

        # calculate cosines with inclination
        cos_i = torch.cos(inclination)
        cos_i2 = cos_i**2
        sin_i = torch.sin(inclination)

        # Precompute exponent and phase terms
        exp_term = torch.exp(-pi * frequency * self.times / quality)
        phase_term = 2 * pi * frequency * self.times + phase

        a_plus = (amplitude / distance) * (1 + cos_i2) * exp_term
        a_cross = (amplitude / distance) * (2 * sin_i) * exp_term

        h_plus = a_plus * torch.cos(phase_term)
        h_cross = a_cross * torch.sin(phase_term)

        # ensure the dtype is double
        h_plus = h_plus.double()
        h_cross = h_cross.double()

        return h_cross, h_plus



from typing import Optional, Tuple

import torch
from jaxtyping import Float
from torch import Tensor

from ...utils.slicing import unfold_windows


class OnlineAverager(torch.nn.Module):
    """
    Module for performing stateful online averaging of
    batches of overlapping timeseries. At present, the
    first `num_updates` predictions produced by this
    model will underestimate the true average.

    Args:
        update_size:
            The number of samples separating the timestamps
            of subsequent inputs.
        batch_size:
            The number of batched inputs to expect at inference
            time.
        num_updates:
            The number of steps over which to average predictions
            before returning them.
        num_channels:
            The expected channel dimension of the input passed
            to the module at inference time.
        offset:
            Number of samples to throw away from the front
            edge of the kernel when averaging.
    """

    def __init__(
        self,
        update_size: int,
        batch_size: int,
        num_updates: int,
        num_channels: int,
        offset: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.update_size = update_size
        self.num_updates = num_updates
        self.batch_size = batch_size
        self.num_channels = num_channels
        self.offset = offset

        # build a blank tensor into which we will embed
        # the updated snapshot predictions at the
        # appropriate time offset for in-batch averaging
        self.batch_update_size = int(batch_size * update_size)
        self.state_size = int((num_updates - 1) * update_size)
        blank_size = self.batch_update_size + self.state_size
        blank = torch.zeros((batch_size, num_channels, blank_size))
        self.register_buffer("blank", blank)

        # set up the indices at which the updated snapshots
        # will be embedded into the blank tensor
        idx = torch.arange(num_updates * update_size)
        idx = torch.stack([idx + i * update_size for i in range(batch_size)])
        idx = idx.view(batch_size, 1, -1).repeat(1, num_channels, 1)
        self.register_buffer("idx", idx)

        # normalization indices used to downweight the
        # existing average at each in-batch aggregation
        weights = torch.scatter(blank, -1, idx, 1).sum(0)
        weight_size = int(num_updates * update_size)
        weights = unfold_windows(weights, weight_size, update_size)
        self.register_buffer("weights", weights)

    def get_initial_state(self) -> Float[Tensor, "channel time"]:
        return torch.zeros((self.num_channels, self.state_size))

    def forward(
        self,
        update: Float[Tensor, "batch channel time1"],
        state: Optional[Float[Tensor, "channel time2"]] = None,
    ) -> Tuple[Float[Tensor, "channel time3"], Float[Tensor, "channel time4"]]:
        if state is None:
            state = self.get_initial_state()

        # slice off the steps from this update closest
        # to the future that we'll actually use. Divide
        # these values by the number of updates up-front
        # for averaging purposes.
        start = -self.num_updates * self.update_size
        if self.offset is not None:
            end = -self.offset
            start += end
        else:
            end = None
        x = update[:, :, start:end] / self.num_updates

        # append zeros to the state into which we
        # can insert our updates
        state = torch.nn.functional.pad(state, (0, self.batch_update_size))

        # window the existing snapshot into overlapping
        # segments and average them with our new updates
        windowed = unfold_windows(state, x.size(-1), self.update_size)
        windowed /= self.weights
        windowed += x

        # embed these windowed averages into a blank
        # array with offsets so that we can add the
        # overlapping bits
        padded = torch.scatter(self.blank, -1, self.idx, windowed)
        new_state = padded.sum(axis=0)

        if self.num_updates == 1:
            # if we don't need stateful behavior,
            # just return the "snapshot" as-is
            output, new_state = new_state, self.get_initial_state()
        else:
            # otherwise split off the values that have finished
            # averaging and are being returned from the ones that
            # will comprise the snapshot at the next update
            splits = [self.batch_size, self.num_updates - 1]
            splits = [i * self.update_size for i in splits]
            output, new_state = torch.split(new_state, splits, dim=-1)
        return output, new_state



import torch
from jaxtyping import Float

from ...constants import MPC_SEC, MTSUN_SI, PI
from ...constants import EulerGamma as GAMMA
from ...types import BatchTensor, FrequencySeries1d
from ..conversion import chirp_mass_and_mass_ratio_to_components


class TaylorF2(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(
        self,
        f: FrequencySeries1d,
        chirp_mass: BatchTensor,
        mass_ratio: BatchTensor,
        chi1: BatchTensor,
        chi2: BatchTensor,
        distance: BatchTensor,
        phic: BatchTensor,
        inclination: BatchTensor,
        f_ref: float,
        **kwargs
    ):
        """
        TaylorF2 up to 3.5 PN in phase. Newtonian SPA amplitude.

        Args:
            f:
                Frequency series in Hz.
            chirp_mass:
                Chirp mass in solar masses
            mass_ratio:
                Mass ratio m1/m2
            chi1:
                Spin of m1
            chi2:
                Spin of m2
            distance:
                Luminosity distance
            phic:
                Phase at coalescence
            inclination:
                Inclination angle
            f_ref:
                Reference frequency

        Returns:
            hc, hp: Tuple[torch.Tensor, torch.Tensor]
                Cross and plus polarizations
        """

        # shape assumed (n_batch, params)
        if (
            chirp_mass.shape[0] != mass_ratio.shape[0]
            or chirp_mass.shape[0] != chi1.shape[0]
            or chi1.shape[0] != chi2.shape[0]
            or chi2.shape[0] != distance.shape[0]
            or distance.shape[0] != phic.shape[0]
            or phic.shape[0] != inclination.shape[0]
        ):
            raise RuntimeError("Tensors should have same batch size")

        mass1, mass2 = chirp_mass_and_mass_ratio_to_components(
            chirp_mass, mass_ratio
        )

        cfac = torch.cos(inclination)
        pfac = 0.5 * (1.0 + cfac * cfac)

        htilde = self.taylorf2_htilde(
            f, mass1, mass2, chi1, chi2, distance, phic, f_ref
        )

        hp = (htilde.mT * pfac).mT
        hc = -1j * (htilde.mT * cfac).mT

        return hc, hp

    def taylorf2_htilde(
        self,
        f: FrequencySeries1d,
        mass1: BatchTensor,
        mass2: BatchTensor,
        chi1: BatchTensor,
        chi2: BatchTensor,
        distance: BatchTensor,
        phic: BatchTensor,
        f_ref: float,
    ) -> Float[FrequencySeries1d, " batch"]:
        mass1_s = mass1 * MTSUN_SI
        mass2_s = mass2 * MTSUN_SI
        M_s = mass1_s + mass2_s
        eta = mass1_s * mass2_s / M_s / M_s

        Mf = torch.outer(M_s, f)
        Mf_ref = torch.outer(M_s, f_ref * torch.ones_like(f))

        Psi, _ = self.taylorf2_phase(Mf, mass1, mass2, chi1, chi2)
        Psi_ref, _ = self.taylorf2_phase(Mf_ref, mass1, mass2, chi1, chi2)

        Psi = (Psi.mT - 2 * phic).mT
        Psi -= Psi_ref

        amp0 = self.taylorf2_amplitude(Mf, mass1, mass2, eta, distance)
        h0 = amp0 * torch.exp(-1j * (Psi - PI / 4))
        return h0

    def taylorf2_amplitude(
        self,
        Mf: BatchTensor,
        mass1: BatchTensor,
        mass2: BatchTensor,
        eta: BatchTensor,
        distance: BatchTensor,
    ) -> Float[FrequencySeries1d, " batch"]:
        mass1_s = mass1 * MTSUN_SI
        mass2_s = mass2 * MTSUN_SI
        v = (PI * Mf) ** (1.0 / 3.0)
        v10 = v**10

        # Flux and energy coefficient at newtonian
        FTaN = 32.0 * eta * eta / 5.0
        dETaN = 2 * (-eta / 2.0)

        amp0 = -4.0 * mass1_s * mass2_s * (PI / 12.0) ** 0.5

        amp0 /= distance * MPC_SEC
        flux = (v10.mT * FTaN).mT
        dEnergy = (v.mT * dETaN).mT
        amp = torch.sqrt(-dEnergy / flux) * v
        amp = (amp.mT * amp0).mT

        return amp

    def taylorf2_phase(
        self,
        Mf: BatchTensor,
        mass1: BatchTensor,
        mass2: BatchTensor,
        chi1: BatchTensor,
        chi2: BatchTensor,
    ) -> Float[FrequencySeries1d, " batch"]:
        """
        Calculate the inspiral phase for the TaylorF2.
        """
        M = mass1 + mass2
        eta = mass1 * mass2 / M / M
        m1byM = mass1 / M
        m2byM = mass2 / M
        chi1sq = chi1 * chi1
        chi2sq = chi2 * chi2

        v0 = torch.ones_like(Mf)
        v1 = (PI * Mf) ** (1.0 / 3.0)
        v2 = v1 * v1
        v3 = v2 * v1
        v4 = v3 * v1
        v5 = v4 * v1
        v6 = v5 * v1
        v7 = v6 * v1
        logv = torch.log(v1)
        v5_logv = v5 * logv
        v6_logv = v6 * logv

        # Phase coeffeciencts from https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c  # noqa E501
        pfaN = 3.0 / (128.0 * eta)
        pfa_v0 = 1.0
        pfa_v1 = 0.0
        pfa_v2 = 5.0 * (74.3 / 8.4 + 11.0 * eta) / 9.0
        pfa_v3 = -16.0 * PI
        # SO contributions at 1.5 PN
        pfa_v3 += (
            m1byM * (25.0 + 38.0 / 3.0 * m1byM) * chi1
            + m2byM * (25.0 + 38.0 / 3.0 * m2byM) * chi2
        )
        pfa_v4 = (
            5.0
            * (3058.673 / 7.056 + 5429.0 / 7.0 * eta + 617.0 * eta * eta)
            / 72.0
        )
        # SO, SS, S1,2-squared contributions
        pfa_v4 += (
            247.0 / 4.8 * eta * chi1 * chi2
            + -721.0 / 4.8 * eta * chi1 * chi2
            + (-720.0 / 9.6 * m1byM * m1byM + 1.0 / 9.6 * m1byM * m1byM)
            * chi1sq
            + (-720.0 / 9.6 * m2byM * m2byM + 1.0 / 9.6 * m2byM * m2byM)
            * chi2sq
            + (240.0 / 9.6 * m1byM * m1byM + -7.0 / 9.6 * m1byM * m1byM)
            * chi1sq
            + (240.0 / 9.6 * m2byM * m2byM + -7.0 / 9.6 * m2byM * m2byM)
            * chi2sq
        )
        pfa_v5logv = 5.0 / 3.0 * (772.9 / 8.4 - 13.0 * eta) * PI
        pfa_v5 = 5.0 / 9.0 * (772.9 / 8.4 - 13.0 * eta) * PI
        # SO coefficient for 2.5 PN
        pfa_v5logv += 3.0 * (
            -m1byM
            * (
                1391.5 / 8.4
                - 10.0 / 3.0 * m1byM * (1.0 - m1byM)
                + m1byM * (1276.0 / 8.1 + 170.0 / 9.0 * m1byM * (1.0 - m1byM))
            )
            * chi1
            - m2byM
            * (
                1391.5 / 8.4
                - 10.0 / 3.0 * m2byM * (1.0 - m2byM)
                + m2byM * (1276.0 / 8.1 + 170.0 / 9.0 * m2byM * (1.0 - m2byM))
            )
            * chi2
        )
        pfa_v5 += (
            -m1byM
            * (
                1391.5 / 8.4
                - 10.0 / 3.0 * m1byM * (1.0 - m1byM)
                + m1byM * (1276.0 / 8.1 + 170.0 / 9.0 * m1byM * (1.0 - m1byM))
            )
            * chi1
            + -m2byM
            * (
                1391.5 / 8.4
                - 10.0 / 3.0 * m2byM * (1.0 - m2byM)
                + m2byM * (1276.0 / 8.1 + 170.0 / 9.0 * m2byM * (1.0 - m2byM))
            )
            * chi2
        )
        pfa_v6logv = -684.8 / 2.1
        pfa_v6 = (
            11583.231236531 / 4.694215680
            - 640.0 / 3.0 * PI * PI
            - 684.8 / 2.1 * GAMMA
            + eta * (-15737.765635 / 3.048192 + 225.5 / 1.2 * PI * PI)
            + eta * eta * 76.055 / 1.728
            - eta * eta * eta * 127.825 / 1.296
            + pfa_v6logv * torch.log(torch.tensor(4.0))
        )
        # SO + S1-S2 + S-squared contribution at 3 PN
        pfa_v6 += (
            PI * m1byM * (1490.0 / 3.0 + m1byM * 260.0) * chi1
            + PI * m2byM * (1490.0 / 3.0 + m2byM * 260.0) * chi2
            + (326.75 / 1.12 + 557.5 / 1.8 * eta) * eta * chi1 * chi2
            + (
                (4703.5 / 8.4 + 2935.0 / 6.0 * m1byM - 120.0 * m1byM * m1byM)
                * m1byM
                * m1byM
                + (
                    -4108.25 / 6.72
                    - 108.5 / 1.2 * m1byM
                    + 125.5 / 3.6 * m1byM * m1byM
                )
                * m1byM
                * m1byM
            )
            * chi1sq
            + (
                (4703.5 / 8.4 + 2935.0 / 6.0 * m2byM - 120.0 * m2byM * m2byM)
                * m2byM
                * m2byM
                + (
                    -4108.25 / 6.72
                    - 108.5 / 1.2 * m2byM
                    + 125.5 / 3.6 * m2byM * m2byM
                )
                * m2byM
                * m2byM
            )
            * chi2sq
        )
        pfa_v7 = PI * (
            770.96675 / 2.54016
            + 378.515 / 1.512 * eta
            - 740.45 / 7.56 * eta * eta
        )
        # SO contribution at 3.5 PN
        pfa_v7 += (
            m1byM
            * (
                -17097.8035 / 4.8384
                + eta * 28764.25 / 6.72
                + eta * eta * 47.35 / 1.44
                + m1byM
                * (
                    -7189.233785 / 1.524096
                    + eta * 458.555 / 3.024
                    - eta * eta * 534.5 / 7.2
                )
            )
        ) * chi1 + (
            m2byM
            * (
                -17097.8035 / 4.8384
                + eta * 28764.25 / 6.72
                + eta * eta * 47.35 / 1.44
                + m2byM
                * (
                    -7189.233785 / 1.524096
                    + eta * 458.555 / 3.024
                    - eta * eta * 534.5 / 7.2
                )
            )
        ) * chi2
        # construct power series
        phasing = (v7.mT * pfa_v7).mT
        phasing += (v6.mT * pfa_v6 + v6_logv.mT * pfa_v6logv).mT
        phasing += (v5.mT * pfa_v5 + v5_logv.mT * pfa_v5logv).mT
        phasing += (v4.mT * pfa_v4).mT
        phasing += (v3.mT * pfa_v3).mT
        phasing += (v2.mT * pfa_v2).mT
        phasing += (v1.mT * pfa_v1).mT
        phasing += (v0.mT * pfa_v0).mT
        # Divide by 0PN v-dependence
        phasing /= v5
        # Multiply by 0PN coefficient
        phasing = (phasing.mT * pfaN).mT

        # Derivative of phase w.r.t Mf
        # dPhi/dMf = dPhi/dv dv/dMf
        Dphasing = (2.0 * v7.mT * pfa_v7).mT
        Dphasing += (v6.mT * (pfa_v6 + pfa_v6logv)).mT
        Dphasing += (v6_logv.mT * pfa_v6logv).mT
        Dphasing += (v5.mT * pfa_v5logv).mT
        Dphasing += (-1.0 * v4.mT * pfa_v4).mT
        Dphasing += (-2.0 * v3.mT * pfa_v3).mT
        Dphasing += (-3.0 * v2.mT * pfa_v2).mT
        Dphasing += (-4.0 * v1.mT * pfa_v1).mT
        Dphasing += -5.0 * v0
        Dphasing /= 3.0 * v1 * v7
        Dphasing *= PI
        Dphasing = (Dphasing.mT * pfaN).mT

        return phasing, Dphasing



import warnings
from typing import Dict, List

import torch
import torch.nn.functional as F
from jaxtyping import Float
from torch import Tensor
from torchaudio.transforms import Spectrogram

from ..types import TimeSeries3d


class MultiResolutionSpectrogram(torch.nn.Module):
    """
    Create a batch of multi-resolution spectrograms
    from a batch of timeseries. Input is expected to
    have the shape `(B, C, T)`, where `B` is the number
    of batches, `C` is the number of channels, and `T`
    is the number of time samples.

    For each timeseries, calculate multiple normalized
    spectrograms based on the `Spectrogram` `kwargs` given.
    Combine the spectrograms by taking the maximum value
    from the nearest time-frequncy bin.

    If the largest number of time bins among the spectrograms
    is `N` and the largest number of frequency bins is `M`,
    the output will have dimensions `(B, C, M, N)`

    Args:
        kernel_length:
            The length in seconds of the time dimension
            of the tensor that will be turned into a
            spectrogram
        sample_rate:
            The sample rate of the timeseries in Hz
        kwargs:
            Arguments passed in kwargs will used to create
            `torchaudio.transforms.Spectrogram`s. Each
            argument should be a list of values. Any list
            of length greater than 1 should be the same
            length
    """

    def __init__(
        self, kernel_length: float, sample_rate: float, **kwargs
    ) -> None:
        super().__init__()
        self.kernel_size = kernel_length * sample_rate
        # This method of combination makes sense only when
        # the spectrograms are normalized, so enforce this
        if "normalized" in kwargs.keys():
            if not all(kwargs["normalized"]):
                raise ValueError(
                    "Received a value of False for 'normalized'. "
                    "This method of combination is sensible only for "
                    "normalized spectrograms."
                )
        else:
            kwargs["normalized"] = [True]
        self.kwargs = self._check_and_format_kwargs(kwargs)

        self.transforms = torch.nn.ModuleList(
            [Spectrogram(**k) for k in self.kwargs]
        )

        dummy_input = torch.ones(int(kernel_length * sample_rate))
        self.shapes = torch.tensor(
            [t(dummy_input).shape for t in self.transforms]
        )

        self.num_freqs = max([shape[0] for shape in self.shapes])
        self.num_times = max([shape[1] for shape in self.shapes])

        left_pad = torch.zeros(len(self.transforms), dtype=torch.int)
        top_pad = torch.zeros(len(self.transforms), dtype=torch.int)
        bottom_pad = torch.tensor(
            [int(self.num_freqs - shape[0]) for shape in self.shapes]
        )
        right_pad = torch.tensor(
            [int(self.num_times - shape[1]) for shape in self.shapes]
        )
        self.register_buffer("left_pad", left_pad)
        self.register_buffer("top_pad", top_pad)
        self.register_buffer("bottom_pad", bottom_pad)
        self.register_buffer("right_pad", right_pad)

        freq_idxs = torch.tensor(
            [
                [int(i * shape[0] / self.num_freqs) for shape in self.shapes]
                for i in range(self.num_freqs)
            ]
        )
        freq_idxs = freq_idxs.repeat(self.num_times, 1, 1).transpose(0, 1)
        time_idxs = torch.tensor(
            [
                [int(i * shape[1] / self.num_times) for shape in self.shapes]
                for i in range(self.num_times)
            ]
        )
        time_idxs = time_idxs.repeat(self.num_freqs, 1, 1)

        self.register_buffer("freq_idxs", freq_idxs)
        self.register_buffer("time_idxs", time_idxs)

    def _check_and_format_kwargs(self, kwargs: Dict[str, List]) -> List:
        lengths = sorted(set([len(v) for v in kwargs.values()]))

        if lengths[-1] > 3:
            warnings.warn(
                "Combining too many spectrograms can impede computation time. "
                "If performance is slower than desired, try reducing the "
                "number of spectrograms",
                RuntimeWarning,
            )

        if len(lengths) > 2 or (len(lengths) == 2 and lengths[0] != 1):
            raise ValueError(
                "Spectrogram keyword args should all have the same "
                f"length or be of length one. Got lengths {lengths}"
            )

        if len(lengths) == 2:
            size = lengths[1]
            kwargs = {k: v * int(size / len(v)) for k, v in kwargs.items()}

        return [dict(zip(kwargs, col)) for col in zip(*kwargs.values())]

    def forward(
        self, X: TimeSeries3d
    ) -> Float[Tensor, "batch channel frequency time"]:
        """
        Calculate spectrograms of the input tensor and
        combine them into a single spectrogram

        Args:
            X:
                Batch of multichannel timeseries which will
                be used to calculate the multi-resolution
                spectrogram. Should have the shape
                `(B, C, T)`, where `B` is the number of
                batches, `C` is the  number of channels,
                and `T` is the number of time samples.
        """
        if X.shape[-1] != self.kernel_size:
            raise ValueError(
                "Expected time dimension to be "
                f"{self.kernel_size} samples long, got input with "
                f"{X.shape[-1]} samples"
            )

        spectrograms = [t(X) for t in self.transforms]

        padded_specs = []
        for spec, left, right, top, bottom in zip(
            spectrograms,
            self.left_pad,
            self.right_pad,
            self.top_pad,
            self.bottom_pad,
        ):
            padded_specs.append(F.pad(spec, (left, right, top, bottom)))

        padded_specs = torch.stack(padded_specs)
        remapped_specs = padded_specs[..., self.freq_idxs, self.time_idxs]
        remapped_specs = torch.diagonal(remapped_specs, dim1=0, dim2=-1)

        return torch.max(remapped_specs, axis=-1)[0]



from typing import Optional

import torch
from jaxtyping import Float
from torch import Tensor

from ..spectral import fast_spectral_density, spectral_density
from ..types import FrequencySeries1to3d, TimeSeries1to3d


class SpectralDensity(torch.nn.Module):
    """
    Transform for computing either the power spectral density
    of a batch of multichannel timeseries, or the cross spectral
    density of two batches of multichannel timeseries.

    On `SpectralDensity.forward` call, if only one tensor is provided,
    this transform will compute its power spectral density. If a second
    tensor is provided, the cross spectral density between the two
    timeseries will be computed. For information about the allowed
    relationships between these two tensors, see the documentation to
    `ml4gw.spectral.fast_spectral_density`.

    Note that the cross spectral density computation is currently
    only available for the `fast_spectral_density` option. If
    `fast=False` and a second tensor is passed to `SpectralDensity.forward`,
    a `NotImplementedError` will be raised.

    Args:
        sample_rate:
            Rate at which tensors passed to `forward` will be sampled
        fftlength:
            Length of the window, in seconds, to use for FFT estimates
        overlap:
            Overlap between windows used for FFT calculation. If left
            as `None`, this will be set to `fftlength / 2`.
        average:
            Aggregation method to use for combining windowed FFTs.
            Allowed values are `"mean"` and `"median"`.
        window:
            Window array to multiply by each FFT window before
            FFT computation. Should have length `nperseg`.
            Defaults to a hanning window.
        fast:
            Whether to use a faster spectral density computation that
            support cross spectral density, or a slower one which does
            not. The cost of the fast implementation is that it is not
            exact for the two lowest frequency bins.
    """

    def __init__(
        self,
        sample_rate: float,
        fftlength: float,
        overlap: Optional[float] = None,
        average: str = "mean",
        window: Optional[
            Float[Tensor, " {int(fftlength*sample_rate)}"]
        ] = None,
        fast: bool = False,
    ) -> None:
        if overlap is None:
            overlap = fftlength / 2
        elif overlap >= fftlength:
            raise ValueError(
                "Can't have overlap {} longer than fftlength {}".format(
                    overlap, fftlength
                )
            )

        super().__init__()

        self.nperseg = int(fftlength * sample_rate)
        self.nstride = self.nperseg - int(overlap * sample_rate)

        # if no window is provided, default to a hanning window;
        # validate that window is correct size
        if window is None:
            window = torch.hann_window(self.nperseg)

        if window.size(0) != self.nperseg:
            raise ValueError(
                "Window must have length {} got {}".format(
                    self.nperseg, window.size(0)
                )
            )
        self.register_buffer("window", window)

        # scale corresponds to "density" normalization, worth
        # considering adding this as a kwarg and changing this calc
        scale = 1.0 / (sample_rate * (self.window**2).sum())
        self.register_buffer("scale", scale)

        if average not in ("mean", "median"):
            raise ValueError(
                f'average must be "mean" or "median", got {average} instead'
            )
        self.average = average
        self.fast = fast

    def forward(
        self, x: TimeSeries1to3d, y: Optional[TimeSeries1to3d] = None
    ) -> FrequencySeries1to3d:
        if self.fast:
            return fast_spectral_density(
                x,
                y=y,
                nperseg=self.nperseg,
                nstride=self.nstride,
                window=self.window,
                scale=self.scale,
                average=self.average,
            )

        if y is not None:
            raise NotImplementedError
        return spectral_density(
            x,
            nperseg=self.nperseg,
            nstride=self.nstride,
            window=self.window,
            scale=self.scale,
            average=self.average,
        )



import math
from typing import Callable, Tuple

import numpy as np
import torch
from jaxtyping import Float
from torch import Tensor

from ..constants import MSUN
from ..transforms import IIRFilter
from ..types import BatchTensor
from .cbc import utils

EXTRA_TIME_FRACTION = (
    0.1  # fraction of waveform duration to add as extra time for tapering
)
EXTRA_CYCLES = 3.0


class TimeDomainCBCWaveformGenerator(torch.nn.Module):
    """
    Waveform generator that generates time-domain waveforms from frequency-domain approximants.

    Frequency domain waveforms are conditioned as done by lalsimulation.
    Specifically, waveforms are generated with a starting frequency `fstart`
    slightly below the requested `f_min`, so that they can be tapered from
    `fstart` to `f_min` using a cosine window.

    Please see https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/group___l_a_l_sim_inspiral__c.html#gac9f16dab2cbca5a431738ee7d2505969 # noqa
    for more information

    Args:
        approximant:
            A callable that returns hplus and hcross polarizations
            given requested frequencies and relevant set of parameters.
            See `ml4gw.waveforms.cbc` for implemented approximants.
        sample_rate:
            Rate at which returned time domain waveform will be
            sampled in Hz. This also specifies `f_max` for generating
            waveforms via the nyquist frequency: `f_max = sample_rate // 2`.
        f_min:
            Lower frequency bound for waveforms
        duration:
            Length of waveform in seconds.
            Waveforms will be left padded with zeros
            appropiately to fill the requested duration
        right_pad:
            How far from the right edge of the window
            in seconds the returned waveform coalescence
            will be placed.
        f_ref:
            Reference frequency for the waveform
    """

    def __init__(
        self,
        approximant: Callable,
        sample_rate: float,
        duration: float,
        f_min: float,
        f_ref: float,
        right_pad: float,
    ) -> None:

        super().__init__()
        self.approximant = approximant
        self.f_min = f_min
        self.sample_rate = sample_rate
        self.duration = duration
        self.right_pad = right_pad
        self.f_ref = f_ref

        self.highpass = self.build_highpass_filter()

    @property
    def delta_t(self):
        return 1 / self.sample_rate

    @property
    def nyquist(self):
        return int(self.sample_rate / 2)

    @property
    def size(self):
        """Number of samples in the waveform"""
        return int(self.duration * self.sample_rate)

    @property
    def delta_f(self):
        return 1 / self.duration

    def build_highpass_filter(self):
        """
        Builds highpass filter object.

        See https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/python/lalsimulation/gwsignal/core/conditioning_subroutines.py?ref_type=heads#L10 # noqa
        """
        order = 8.0
        w1 = np.tan(np.pi * (self.f_min) / self.sample_rate)
        attenuation = 0.99
        wc = w1 * (1.0 / attenuation**0.5 - 1) ** (1.0 / (2.0 * order))
        fc = self.sample_rate * np.arctan(wc) / np.pi

        return IIRFilter(
            order,
            fc / (self.sample_rate / 2),
            btype="highpass",
            ftype="butterworth",
        )

    def get_frequencies(self, df: float):
        """Get the frequencies from 0 to nyquist for corresponding df"""
        num_freqs = int(self.nyquist / df) + 1
        return torch.linspace(0, self.nyquist, num_freqs)

    def generate_conditioned_fd_waveform(
        self, **parameters: dict[str, BatchTensor]
    ) -> Tuple[Float[Tensor, "{N} samples"], Float[Tensor, "{N} samples"]]:
        """
        Generate a conditioned frequency domain waveform from a frequency domain approximant.

        Based on https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/python/lalsimulation/gwsignal/core/waveform_conditioning.py?ref_type=heads#L248 # noqa

        Args:
            **parameters:
                Dictionary of parameters for waveform generation
                where key is the parameter name and value is a tensor of parameters.
                It is required that `parameters` contains `mass_1`, `mass_2`, `s1z`, and `s2z`
                keys, which are used for determining parameters of data conditioning.
                If the specified approximant takes other parameters for waveform generation,
                like `chirp_mass` and `mass_ratio`, the utility functions in `ml4gw.waveforms.conversion`
                may be useful for populating the parameters dictionary with these additional parameters.
                Note that, if using an approximant from `ml4gw.waveforms.cbc`, any additional keys in `parameters`
                not ingested by the approximant will be ignored.
        """
        # convert masses to kg, make sure
        # they are doubles so there is no
        # overflow in the calculations
        mass_1, mass_2 = (
            parameters["mass_1"].double() * MSUN,
            parameters["mass_2"].double() * MSUN,
        )
        total_mass = mass_1 + mass_2
        s1z, s2z = parameters["s1z"], parameters["s2z"]
        device = mass_1.device

        f_isco = utils.frequency_isco(mass_1, mass_2)
        f_min = torch.minimum(
            f_isco,
            torch.tensor(self.f_min, device=device),
        )

        # upper bound on chirp time
        tchirp = utils.chirp_time_bound(f_min, mass_1, mass_2, s1z, s2z)

        # upper bound on final black hole spin
        s = utils.final_black_hole_spin_bound(s1z, s2z)

        # upper bound on the final plunge, merger, and ringdown time
        tmerge = utils.merge_time_bound(
            mass_1, mass_2
        ) + utils.ringdown_time_bound(total_mass, s)

        # extra time to include for all waveforms to take care of situations
        # where the frequency is close to merger (and is sweeping rapidly):
        # this is a few cycles at the low frequency
        textra = EXTRA_CYCLES / f_min

        # lower bound on chirpt frequency start used for
        # conditioning the frequency domain waveform
        fstart = utils.chirp_start_frequency_bound(
            (1.0 + EXTRA_TIME_FRACTION) * tchirp, mass_1, mass_2
        )

        # revised chirp time estimate based on fstart
        tchirp_fstart = utils.chirp_time_bound(
            fstart, mass_1, mass_2, s1z, s2z
        )

        # chirp length in samples
        chirplen = torch.round(
            (tchirp_fstart + tmerge + 2.0 * textra) * self.sample_rate
        )

        # pad to next power of 2
        chirplen = 2 ** torch.ceil(torch.log(chirplen) / math.log(2))

        # get smallest df corresponding to longest chirp length,
        # which will make sure there is no wrap around effects.
        df = 1.0 / (chirplen.max() / self.sample_rate)

        # generate frequency array from 0 to nyquist based on df
        frequencies = self.get_frequencies(df).to(mass_1.device)

        # downselect to frequencies above fstart,
        # and generate the waveform at the specified frequencies
        freq_mask = frequencies >= fstart.min()
        waveform_frequencies = frequencies[freq_mask]

        # generate the waveform at specified frequencies
        cross, plus = self.approximant(
            waveform_frequencies, **parameters, f_ref=self.f_ref
        )
        batch_size = cross.size(0)

        # create tensors to hold the full spectrum
        # of frequencies from 0 to nyquist, and then
        # fill in the requested frequencies with the waveform values
        shape = (batch_size, frequencies.size(0))
        hc_spectrum = torch.zeros(shape, dtype=cross.dtype, device=device)
        hp_spectrum = torch.zeros(shape, dtype=plus.dtype, device=device)

        hc_spectrum[:, freq_mask] = cross
        hp_spectrum[:, freq_mask] = plus

        # build a taper that is dependent on each
        # individual waveforms fstart;
        # since this means that the taper sizes
        # will be different for each waveform,
        # construct the tapers based on the maximum size
        # and then set the values outside of the individual
        # waveform taper regions to 1.0
        k0s = torch.round(fstart / df)
        k1s = torch.round(f_min / df)

        num_freqs = frequencies.size(0)
        frequency_indices = torch.arange(num_freqs, device=device)
        taper_mask = frequency_indices <= k1s[:, None]
        taper_mask &= frequency_indices >= k0s[:, None]

        indices = frequency_indices.expand(batch_size, -1)

        kvals = indices[taper_mask]
        k0s_expanded = k0s.unsqueeze(1).expand(-1, num_freqs)[taper_mask]
        k1s_expanded = k1s.unsqueeze(1).expand(-1, num_freqs)[taper_mask]

        windows = 0.5 - 0.5 * torch.cos(
            torch.pi * (kvals - k0s_expanded) / (k1s_expanded - k0s_expanded)
        )

        hc_spectrum[taper_mask] *= windows
        hp_spectrum[taper_mask] *= windows

        # zero out frequencies below fstart
        zero_mask = frequencies < fstart[:, None]
        hc_spectrum[zero_mask] = 0
        hp_spectrum[zero_mask] = 0

        # set nyquist frequency to zero
        hc_spectrum[..., -1], hp_spectrum[..., -1] = 0.0, 0.0

        # apply time translation in (i.e. phase shift in frequency domain)
        # that will translate the coalescense time such that it is `right_pad`
        # seconds from the right edge of the window
        tshift = round(self.right_pad * self.sample_rate) / self.sample_rate
        kvals = torch.arange(num_freqs, device=device)
        phase_shift = torch.exp(1j * 2 * torch.pi * df * tshift * kvals)

        hc_spectrum *= phase_shift
        hp_spectrum *= phase_shift

        return hc_spectrum, hp_spectrum

    def forward(
        self,
        **parameters,
    ) -> Tuple[Float[Tensor, "{N} samples"], Float[Tensor, "{N} samples"]]:
        """
        Generates a time-domain waveform from a frequency domain approximant.
        Conditioning is based onhttps://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/python/lalsimulation/gwsignal/core/waveform_conditioning.py?ref_type=heads#L248 # noqa

        A frequency domain waveform is generated, conditioned (see `generate_conditioned_fd_waveform`)
        and fftdd into the time-domain

        **parameters:
            Dictionary of parameters for waveform generation
            where key is the parameter name and value is a tensor of parameters.
            It is required that `parameters` contains `mass_1`, `mass_2`, `s1z`, and `s2z`
            keys, which are used for determining parameters of data conditioning.
            If the specified approximant takes other parameters for waveform generation,
            like `chirp_mass` and `mass_ratio`, the utility functions in `ml4gw.waveforms.conversion`
            may be useful for populating the parameters dictionary with these additional parameters.
            Note that, if using an approximant from `ml4gw.waveforms.cbc`, any additional keys in `parameters`
            not ingested by the approximant will be ignored.
        """

        hc, hp = self.generate_conditioned_fd_waveform(**parameters)

        # fft to time domain and apply appropiate scaling
        hc = torch.fft.irfft(hc) * self.sample_rate
        hp = torch.fft.irfft(hp) * self.sample_rate

        # TODO: some additional tapering in the time
        # domain is performed in lalsimulation

        # pad waveforms on left up to requested duration
        pad = int((self.duration * self.sample_rate) - hp.shape[-1])
        hc = torch.nn.functional.pad(hc, (pad, 0))
        hp = torch.nn.functional.pad(hp, (pad, 0))

        # finally, highpass the waveforms,
        # going to double precision
        hp = self.highpass(hp.double())
        hc = self.highpass(hc.double())

        return hc, hp



from typing import Callable, Optional

import torch
from jaxtyping import Float
from torch import Tensor

NormLayer = Callable[[int], torch.nn.Module]


class GroupNorm1D(torch.nn.Module):
    """
    Custom implementation of GroupNorm which is faster than the
    out-of-the-box PyTorch version at inference time.
    """

    def __init__(
        self,
        num_channels: int,
        num_groups: Optional[int] = None,
        eps: float = 1e-5,
    ):
        super().__init__()
        num_groups = num_groups or num_channels
        if num_channels % num_groups:
            raise ValueError("num_groups must be a factor of num_channels")

        self.num_channels = num_channels
        self.num_groups = num_groups
        self.channels_per_group = self.num_channels // self.num_groups
        self.eps = eps

        shape = (self.num_channels, 1)
        self.weight = torch.nn.Parameter(torch.ones(shape))
        self.bias = torch.nn.Parameter(torch.zeros(shape))

    def forward(
        self, x: Float[Tensor, "batch channel length"]
    ) -> Float[Tensor, "batch channel length"]:
        if len(x.shape) != 3:
            raise ValueError(
                "GroupNorm1D requires 3-dimensional input, "
                f"received {len(x.shape)} dimensional input"
            )

        keepdims = self.num_groups == self.num_channels

        # compute group variance via the E[x**2] - E**2[x] trick
        mean = x.mean(-1, keepdims=keepdims)
        sq_mean = (x**2).mean(-1, keepdims=keepdims)

        # if we have groups, do some reshape magic
        # to calculate group level stats then
        # reshape back to full channel dimension
        if self.num_groups != self.num_channels:
            mean = torch.stack([mean, sq_mean], dim=1)
            mean = mean.reshape(
                -1, 2, self.num_groups, self.channels_per_group
            )
            mean = mean.mean(-1, keepdims=True)
            mean = mean.expand(-1, -1, -1, self.channels_per_group)
            mean = mean.reshape(-1, 2, self.num_channels, 1)
            mean, sq_mean = mean[:, 0], mean[:, 1]

        # roll the mean and variance into the
        # weight and bias so that we have to do
        # fewer computations along the full time axis
        std = (sq_mean - mean**2 + self.eps) ** 0.5
        scale = self.weight / std
        shift = self.bias - scale * mean
        return shift + x * scale


class GroupNorm1DGetter:
    """
    Utility for making a NormLayer Callable that maps from
    an integer number of channels to a torch Module. Useful
    for command-line parameterization with jsonargparse.
    """

    def __init__(self, groups: Optional[int] = None) -> None:
        self.groups = groups

    def __call__(self, num_channels: int) -> torch.nn.Module:
        if self.groups is None:
            num_groups = None
        else:
            num_groups = min(num_channels, self.groups)
        return GroupNorm1D(num_channels, num_groups)


# TODO generalize faster 1dDGroupNorm to 2D
class GroupNorm2DGetter:
    """
    Utility for making a NormLayer Callable that maps from
    an integer number of channels to a torch Module. Useful
    for command-line parameterization with jsonargparse.
    """

    def __init__(self, groups: Optional[int] = None) -> None:
        self.groups = groups

    def __call__(self, num_channels: int) -> torch.nn.Module:
        if self.groups is None:
            num_groups = num_channels
        else:
            num_groups = min(num_channels, self.groups)
        return torch.nn.GroupNorm(num_groups, num_channels)



"""
Utilities for conditioning waveforms
See https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c # noqa
"""
import torch

from ml4gw.constants import MRSUN, MSUN, MTSUN_SI, C, G
from ml4gw.types import BatchTensor
from ml4gw.waveforms.cbc import coefficients


def chirp_time_bound(
    fstart: BatchTensor,
    mass_1: BatchTensor,
    mass_2: BatchTensor,
    s1: BatchTensor,
    s2: BatchTensor,
) -> BatchTensor:
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L4969
    """

    total_mass = mass_1 + mass_2
    reduced_mass = mass_1 * mass_2 / total_mass
    eta = reduced_mass / total_mass
    chi = torch.max(s1.abs(), s2.abs()).abs()

    c0 = torch.abs(coefficients.taylor_t2_timing_0pn_coeff(total_mass, eta))

    c2 = coefficients.taylor_t2_timing_2pn_coeff(eta)
    c3 = (226.0 / 15.0) * chi
    c4 = coefficients.taylor_t2_timing_4pn_coeff(eta)

    v = (torch.pi * total_mass * fstart * G) ** (1.0 / 3.0)
    v /= C

    return (c0 * (v**-8) * (1 + (c2 + (c3 + c4 * v) * v) * v * v)).float()


def chirp_start_frequency_bound(
    tchirp: BatchTensor,
    mass_1: BatchTensor,
    mass_2: BatchTensor,
):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L5104
    """
    total_mass = mass_1 + mass_2
    mu = mass_1 * mass_2 / total_mass

    eta = mu / total_mass
    c0 = coefficients.taylor_t3_frequency_0pn_coeff(total_mass)
    return (
        c0
        * pow(5.0 * total_mass * (MTSUN_SI / MSUN) / (eta * tchirp), 3.0 / 8.0)
    ).float()


def final_black_hole_spin_bound(
    s1z: BatchTensor, s2z: BatchTensor
) -> BatchTensor:
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L5081
    """
    maximum_black_hole_spin = 0.998
    s = 0.686 + 0.15 * (s1z + s2z)
    s = torch.maximum(s, torch.abs(s1z)).maximum(torch.abs(s2z))
    s = torch.clamp(s, max=maximum_black_hole_spin)
    return s


def merge_time_bound(mass_1: BatchTensor, mass_2: BatchTensor) -> BatchTensor:
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L5007
    """

    n_orbits = 1
    total_mass = mass_1 + mass_2
    r = 9.0 * total_mass * MRSUN / MSUN
    v = C / 3.0
    return (n_orbits * (2.0 * torch.pi * r / v)).float()


def ringdown_time_bound(
    total_mass: BatchTensor, s: BatchTensor
) -> BatchTensor:
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L5032
    """
    n_efolds = 11

    f1 = 1.5251
    f2 = -1.1568
    f3 = 0.1292
    q1 = 0.7000
    q2 = 1.4187
    q3 = -0.4990

    omega = (f1 + f2 * (1.0 - s) ** f3) / (total_mass * MTSUN_SI / MSUN)
    Q = q1 + q2 * (1.0 - s) ** q3
    tau = 2.0 * Q / omega
    return (n_efolds * tau).float()


def frequency_isco(mass_1: BatchTensor, mass_2: BatchTensor):
    return (
        1.0
        / (
            (9.0**1.5) * torch.pi * (mass_1 + mass_2) * MTSUN_SI / MSUN
        ).float()
    )



import torch
from jaxtyping import Float

from ...constants import MTSUN_SI, PI
from ...types import BatchTensor, FrequencySeries1d
from .phenom_d_data import QNMData_a, QNMData_fdamp, QNMData_fring
from .taylorf2 import TaylorF2


class IMRPhenomD(TaylorF2):
    def __init__(self):
        super().__init__()
        self.register_buffer("qnmdata_a", QNMData_a)
        self.register_buffer("qnmdata_fdamp", QNMData_fdamp)
        self.register_buffer("qnmdata_fring", QNMData_fring)

    def forward(
        self,
        f: FrequencySeries1d,
        chirp_mass: BatchTensor,
        mass_ratio: BatchTensor,
        chi1: BatchTensor,
        chi2: BatchTensor,
        distance: BatchTensor,
        phic: BatchTensor,
        inclination: BatchTensor,
        f_ref: float,
        **kwargs
    ):
        """
        IMRPhenomD waveform

        Args:
            f:
                Frequency series in Hz.
            chirp_mass:
                Chirp mass in solar masses
            mass_ratio:
                Mass ratio m1/m2
            chi1:
                Spin of m1
            chi2:
                Spin of m2
            distance:
                Distance to source in Mpc
            phic:
                Phase at coalescence
            inclination:
                Inclination of the source
            f_ref:
                Reference frequency

        Returns:
            hc, hp: Tuple[torch.Tensor, torch.Tensor]
                Cross and plus polarizations
        """
        # shape assumed (n_batch, params)
        if (
            chirp_mass.shape[0] != mass_ratio.shape[0]
            or mass_ratio.shape[0] != chi1.shape[0]
            or chi1.shape[0] != chi2.shape[0]
            or chi2.shape[0] != distance.shape[0]
            or distance.shape[0] != phic.shape[0]
            or phic.shape[0] != inclination.shape[0]
        ):
            raise RuntimeError("Tensors should have same batch size")
        cfac = torch.cos(inclination)
        pfac = 0.5 * (1.0 + cfac * cfac)

        htilde = self.phenom_d_htilde(
            f, chirp_mass, mass_ratio, chi1, chi2, distance, phic, f_ref
        )

        hp = (htilde.mT * pfac).mT
        hc = -1j * (htilde.mT * cfac).mT

        return hc, hp

    def phenom_d_htilde(
        self,
        f: FrequencySeries1d,
        chirp_mass: BatchTensor,
        mass_ratio: BatchTensor,
        chi1: BatchTensor,
        chi2: BatchTensor,
        distance: BatchTensor,
        phic: BatchTensor,
        f_ref: float,
    ) -> Float[FrequencySeries1d, " batch"]:
        total_mass = chirp_mass * (1 + mass_ratio) ** 1.2 / mass_ratio**0.6
        mass_1 = total_mass / (1 + mass_ratio)
        mass_2 = mass_1 * mass_ratio
        eta = (chirp_mass / total_mass) ** (5 / 3)
        eta2 = eta * eta
        Seta = torch.sqrt(1.0 - 4.0 * eta)
        chi = self.chiPN(Seta, eta, chi1, chi2)
        chi22 = chi2 * chi2
        chi12 = chi1 * chi1
        xi = -1.0 + chi
        M_s = total_mass * MTSUN_SI

        gamma2 = self.gamma2_fun(eta, eta2, xi)
        gamma3 = self.gamma3_fun(eta, eta2, xi)

        fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)
        Mf_peak = self.fmaxCalc(fRD, fDM, gamma2, gamma3)
        _, t0 = self.phenom_d_mrd_phase(Mf_peak, eta, eta2, chi1, chi2, xi)

        Mf = torch.outer(M_s, f)
        Mf_ref = torch.outer(M_s, f_ref * torch.ones_like(f))

        Psi, _ = self.phenom_d_phase(
            Mf, mass_1, mass_2, eta, eta2, chi1, chi2, xi
        )
        Psi_ref, _ = self.phenom_d_phase(
            Mf_ref, mass_1, mass_2, eta, eta2, chi1, chi2, xi
        )

        Psi = (Psi.mT - 2 * phic).mT
        Psi -= Psi_ref
        Psi -= ((Mf - Mf_ref).mT * t0).mT

        amp, _ = self.phenom_d_amp(
            Mf,
            mass_1,
            mass_2,
            eta,
            eta2,
            Seta,
            chi1,
            chi2,
            chi12,
            chi22,
            xi,
            distance,
        )

        amp_0 = self.taylorf2_amplitude(
            Mf, mass_1, mass_2, eta, distance
        )  # this includes f^(-7/6) dependence

        h0 = -amp_0 * amp * torch.exp(-1j * Psi)

        return h0

    def phenom_d_amp(
        self,
        Mf,
        mass_1,
        mass_2,
        eta,
        eta2,
        Seta,
        chi1,
        chi2,
        chi12,
        chi22,
        xi,
        distance,
        fRD=None,  # used for passing ringdown frequency from phenom_p
        fDM=None,  # used for passing damping frequency from phenom_p
    ):
        ins_amp, ins_Damp = self.phenom_d_inspiral_amp(
            Mf, eta, eta2, Seta, xi, chi1, chi2, chi12, chi22
        )
        int_amp, int_Damp = self.phenom_d_int_amp(
            Mf, eta, eta2, Seta, chi1, chi2, chi12, chi22, xi, fRD, fDM
        )
        mrd_amp, mrd_Damp = self.phenom_d_mrd_amp(
            Mf, eta, eta2, chi1, chi2, xi, fRD, fDM
        )

        gamma2 = self.gamma2_fun(eta, eta2, xi)
        gamma3 = self.gamma3_fun(eta, eta2, xi)

        # merger ringdown
        if (fRD is None) != (fDM is None):
            raise ValueError(
                "Both fRD and fDM must either be provided or both be None"
            )
        if (fRD is None) and (fDM is None):
            fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)

        Mf_peak = self.fmaxCalc(fRD, fDM, gamma2, gamma3)
        # Geometric peak and joining frequencies
        Mf_peak = (torch.ones_like(Mf).mT * Mf_peak).mT
        Mf_join_ins = 0.014 * torch.ones_like(Mf)

        # construct full IMR Amp
        theta_minus_f1 = torch.heaviside(
            Mf_join_ins - Mf, torch.tensor(0.0, device=Mf.device)
        )
        theta_plus_f1 = torch.heaviside(
            Mf - Mf_join_ins, torch.tensor(1.0, device=Mf.device)
        )
        theta_minus_f2 = torch.heaviside(
            Mf_peak - Mf, torch.tensor(0.0, device=Mf.device)
        )
        theta_plus_f2 = torch.heaviside(
            Mf - Mf_peak, torch.tensor(1.0, device=Mf.device)
        )

        amp = theta_minus_f1 * ins_amp
        amp += theta_plus_f1 * int_amp * theta_minus_f2
        amp += theta_plus_f2 * mrd_amp

        Damp = theta_minus_f1 * ins_Damp
        Damp += theta_plus_f1 * int_Damp * theta_minus_f2
        Damp += theta_plus_f2 * mrd_Damp

        return amp, Damp

    def phenom_d_int_amp(
        self,
        Mf,
        eta,
        eta2,
        Seta,
        chi1,
        chi2,
        chi12,
        chi22,
        xi,
        fRD=None,  # used for passing ringdown frequency from phenom_p
        fDM=None,  # used for passing damping frequency from phenom_p
    ):
        # merger ringdown
        if (fRD is None) != (fDM is None):
            raise ValueError(
                "Both fRD and fDM must either be provided or both be None"
            )
        if (fRD is None) and (fDM is None):
            fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)

        # Geometric frequency definition from PhenomD header file
        AMP_fJoin_INS = 0.014

        Mf1 = AMP_fJoin_INS * torch.ones_like(Mf)
        gamma2 = self.gamma2_fun(eta, eta2, xi)
        gamma3 = self.gamma3_fun(eta, eta2, xi)

        fpeak = self.fmaxCalc(fRD, fDM, gamma2, gamma3)
        Mf3 = (torch.ones_like(Mf).mT * fpeak).mT
        dfx = 0.5 * (Mf3 - Mf1)
        Mf2 = Mf1 + dfx

        v1, d1 = self.phenom_d_inspiral_amp(
            Mf1, eta, eta2, Seta, xi, chi1, chi2, chi12, chi22
        )
        v3, d2 = self.phenom_d_mrd_amp(
            Mf3, eta, eta2, chi1, chi2, xi, fRD, fDM
        )
        v2 = (
            torch.ones_like(Mf).mT * self.AmpIntColFitCoeff(eta, eta2, xi)
        ).mT

        delta_0, delta_1, delta_2, delta_3, delta_4 = self.delta_values(
            f1=Mf1, f2=Mf2, f3=Mf3, v1=v1, v2=v2, v3=v3, d1=d1, d2=d2
        )

        amp = (
            delta_0
            + Mf * delta_1
            + Mf**2 * (delta_2 + Mf * delta_3 + Mf**2 * delta_4)
        )
        Damp = delta_1 + Mf * (
            2 * delta_2 + 3 * Mf * delta_3 + 4 * Mf**2 * delta_4
        )
        return amp, Damp

    # fRD and fDM are to be passed for generating phenom_p waveforms
    # and remain None for phenom_d
    def phenom_d_mrd_amp(
        self, Mf, eta, eta2, chi1, chi2, xi, fRD=None, fDM=None
    ):
        # merger ringdown
        if (fRD is None) != (fDM is None):
            raise ValueError(
                "Both fRD and fDM must either be provided or both be None"
            )
        if (fRD is None) and (fDM is None):
            fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)

        gamma1 = self.gamma1_fun(eta, eta2, xi)
        gamma2 = self.gamma2_fun(eta, eta2, xi)
        gamma3 = self.gamma3_fun(eta, eta2, xi)
        fDMgamma3 = fDM * gamma3
        pow2_fDMgamma3 = (torch.ones_like(Mf).mT * fDMgamma3 * fDMgamma3).mT
        fminfRD = Mf - (torch.ones_like(Mf).mT * fRD).mT
        exp_times_lorentzian = torch.exp(fminfRD.mT * gamma2 / fDMgamma3).mT
        exp_times_lorentzian *= fminfRD**2 + pow2_fDMgamma3

        amp = (1 / exp_times_lorentzian.mT * gamma1 * gamma3 * fDM).mT
        Damp = (fminfRD.mT * -2 * fDM * gamma1 * gamma3) / (
            fminfRD * fminfRD + pow2_fDMgamma3
        ).mT - (gamma2 * gamma1)
        Damp = Damp.mT / exp_times_lorentzian
        return amp, Damp

    def phenom_d_inspiral_amp(
        self, Mf, eta, eta2, Seta, xi, chi1, chi2, chi12, chi22
    ):
        SetaPlus1 = 1 + Seta

        Mf_one_third = Mf ** (1.0 / 3.0)
        Mf_two_third = Mf_one_third * Mf_one_third
        Mf_four_third = Mf_two_third * Mf_two_third
        Mf_five_third = Mf_four_third * Mf_one_third
        Mf_seven_third = Mf_five_third * Mf_two_third
        MF_eight_third = Mf_seven_third * Mf_one_third
        Mf_two = Mf * Mf
        Mf_three = Mf_two * Mf

        prefactors_two_thirds = ((-969 + 1804 * eta) * PI ** (2.0 / 3.0)) / 672
        prefactors_one = (
            (
                chi1 * (81 * SetaPlus1 - 44 * eta)
                + chi2 * (81 - 81 * Seta - 44 * eta)
            )
            * PI
        ) / 48.0
        prefactors_four_thirds = (
            (
                -27312085.0
                - 10287648 * chi22
                - 10287648 * chi12 * SetaPlus1
                + 10287648 * chi22 * Seta
                + 24
                * (
                    -1975055
                    + 857304 * chi12
                    - 994896 * chi1 * chi2
                    + 857304 * chi22
                )
                * eta
                + 35371056 * eta2
            )
            * PI ** (4.0 / 3.0)
        ) / 8.128512e6
        prefactors_five_thirds = (
            PI ** (5.0 / 3.0)
            * (
                chi2
                * (
                    -285197 * (-1 + Seta)
                    + 4 * (-91902 + 1579 * Seta) * eta
                    - 35632 * eta2
                )
                + chi1
                * (
                    285197 * SetaPlus1
                    - 4 * (91902 + 1579 * Seta) * eta
                    - 35632 * eta2
                )
                + 42840 * (-1.0 + 4 * eta) * PI
            )
        ) / 32256.0
        prefactors_two = (
            -(
                PI**2
                * (
                    -336
                    * (
                        -3248849057.0
                        + 2943675504 * chi12
                        - 3339284256 * chi1 * chi2
                        + 2943675504 * chi22
                    )
                    * eta2
                    - 324322727232 * eta2 * eta
                    - 7
                    * (
                        -177520268561
                        + 107414046432 * chi22
                        + 107414046432 * chi12 * SetaPlus1
                        - 107414046432 * chi22 * Seta
                        + 11087290368
                        * (chi1 + chi2 + chi1 * Seta - chi2 * Seta)
                        * PI
                    )
                    + 12
                    * eta
                    * (
                        -545384828789
                        - 176491177632 * chi1 * chi2
                        + 202603761360 * chi22
                        + 77616 * chi12 * (2610335 + 995766 * Seta)
                        - 77287373856 * chi22 * Seta
                        + 5841690624 * (chi1 + chi2) * PI
                        + 21384760320 * PI**2
                    )
                )
            )
            / 6.0085960704e10
        )
        prefactors_seven_thirds = self.rho1_fun(eta, eta2, xi)
        prefactors_eight_thirds = self.rho2_fun(eta, eta2, xi)
        prefactors_three = self.rho3_fun(eta, eta2, xi)

        amp = torch.ones_like(Mf)
        amp += (
            Mf_two_third.mT * prefactors_two_thirds
            + Mf_four_third.mT * prefactors_four_thirds
            + Mf_five_third.mT * prefactors_five_thirds
            + Mf_seven_third.mT * prefactors_seven_thirds
            + MF_eight_third.mT * prefactors_eight_thirds
            + Mf.mT * prefactors_one
            + Mf_two.mT * prefactors_two
            + Mf_three.mT * prefactors_three
        ).mT

        Damp = (
            (2.0 / 3.0) / Mf_one_third.mT * prefactors_two_thirds
            + (4.0 / 3.0) * Mf_one_third.mT * prefactors_four_thirds
            + (5.0 / 3.0) * Mf_two_third.mT * prefactors_five_thirds
            + (7.0 / 3.0) * Mf_four_third.mT * prefactors_seven_thirds
            + (8.0 / 3.0) * Mf_five_third.mT * prefactors_eight_thirds
            + prefactors_one
            + 2.0 * Mf.mT * prefactors_two
            + 3.0 * Mf_two.mT * prefactors_three
        ).mT

        return amp, Damp

    # fRD and fDM are to be passed for generating phenom_p waveforms
    # and remain None for phenom_d
    def phenom_d_phase(
        self, Mf, mass_1, mass_2, eta, eta2, chi1, chi2, xi, fRD=None, fDM=None
    ):
        ins_phase, ins_Dphase = self.phenom_d_inspiral_phase(
            Mf, mass_1, mass_2, eta, eta2, xi, chi1, chi2
        )
        int_phase, int_Dphase = self.phenom_d_int_phase(Mf, eta, eta2, xi)
        mrd_phase, mrd_Dphase = self.phenom_d_mrd_phase(
            Mf, eta, eta2, chi1, chi2, xi, fRD, fDM
        )

        # merger ringdown
        if (fRD is None) != (fDM is None):
            raise ValueError(
                "Both fRD and fDM must either be provided or both be None"
            )
        if (fRD is None) and (fDM is None):
            fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)
        # definitions in Eq. (35) of arXiv:1508.07253
        # PHI_fJoin_INS in header LALSimIMRPhenomD.h
        # C1 continuity at intermediate region i.e. f_1
        PHI_fJoin_INS = 0.018 * torch.ones_like(Mf)
        ins_phase_f1, ins_Dphase_f1 = self.phenom_d_inspiral_phase(
            PHI_fJoin_INS, mass_1, mass_2, eta, eta2, xi, chi1, chi2
        )
        int_phase_f1, int_Dphase_f1 = self.phenom_d_int_phase(
            PHI_fJoin_INS, eta, eta2, xi
        )
        C2Int = ins_Dphase_f1 - int_Dphase_f1
        C1Int = (
            ins_phase_f1 - (int_phase_f1.mT / eta).mT - C2Int * PHI_fJoin_INS
        )
        # C1 continuity at ringdown
        fRDJoin = (0.5 * torch.ones_like(Mf).mT * fRD).mT
        int_phase_rd, int_Dphase_rd = self.phenom_d_int_phase(
            fRDJoin, eta, eta2, xi
        )
        mrd_phase_rd, mrd_Dphase_rd = self.phenom_d_mrd_phase(
            fRDJoin, eta, eta2, chi1, chi2, xi, fRD, fDM
        )
        PhiIntTempVal = (int_phase_rd.mT / eta).mT + C1Int + C2Int * fRDJoin
        # C2MRD = int_Dphase_rd - mrd_Dphase_rd
        C2MRD = C2Int + int_Dphase_rd - mrd_Dphase_rd
        C1MRD = PhiIntTempVal - (mrd_phase_rd.mT / eta).mT - C2MRD * fRDJoin

        int_phase = (int_phase.mT / eta).mT
        int_phase += C1Int
        int_phase += Mf * C2Int

        mrd_phase = (mrd_phase.mT / eta).mT
        mrd_phase += C1MRD
        mrd_phase += Mf * C2MRD

        # construct full IMR phase
        theta_minus_f1 = torch.heaviside(
            PHI_fJoin_INS - Mf, torch.tensor(0.0, device=Mf.device)
        )
        theta_plus_f1 = torch.heaviside(
            Mf - PHI_fJoin_INS, torch.tensor(1.0, device=Mf.device)
        )
        theta_minus_f2 = torch.heaviside(
            fRDJoin - Mf, torch.tensor(0.0, device=Mf.device)
        )
        theta_plus_f2 = torch.heaviside(
            Mf - fRDJoin, torch.tensor(1.0, device=Mf.device)
        )

        phasing = theta_minus_f1 * ins_phase
        phasing += theta_plus_f1 * int_phase * theta_minus_f2
        phasing += theta_plus_f2 * mrd_phase

        Dphasing = theta_minus_f1 * ins_Dphase
        Dphasing += theta_plus_f1 * int_Dphase * theta_minus_f2
        Dphasing += theta_plus_f2 * mrd_Dphase

        return phasing, Dphasing

    # fRD and fDM are to be passed for generating phenom_p waveforms
    # and remain None for phenom_d
    def phenom_d_mrd_phase(
        self, Mf, eta, eta2, chi1, chi2, xi, fRD=None, fDM=None
    ):
        alpha1 = self.alpha1Fit(eta, eta2, xi)
        alpha2 = self.alpha2Fit(eta, eta2, xi)
        alpha3 = self.alpha3Fit(eta, eta2, xi)
        alpha4 = self.alpha4Fit(eta, eta2, xi)
        alpha5 = self.alpha5Fit(eta, eta2, xi)

        # merger ringdown
        if (fRD is None) != (fDM is None):
            raise ValueError(
                "Both fRD and fDM must either be provided or both be None"
            )
        if (fRD is None) and (fDM is None):
            fRD, fDM = self.fring_fdamp(eta, eta2, chi1, chi2)

        f_minus_alpha5_fRD = (Mf.t() - alpha5 * fRD).t()
        # Leading 1/eta is not multiplied at this stage
        mrd_phasing = (Mf.t() * alpha1).t()
        mrd_phasing -= (1 / Mf.t() * alpha2).t()
        mrd_phasing += (4.0 / 3.0) * (Mf.t() ** (3.0 / 4.0) * alpha3).t()
        mrd_phasing += (torch.atan(f_minus_alpha5_fRD.t() / fDM) * alpha4).t()

        mrd_Dphasing = (
            alpha4 * fDM / (f_minus_alpha5_fRD.t() ** 2 + fDM**2)
        ).t()
        mrd_Dphasing += (Mf.t() ** (-1.0 / 4.0) * alpha3).t()
        mrd_Dphasing += (Mf.t() ** (-2.0) * alpha2).t()
        mrd_Dphasing = (mrd_Dphasing.t() + alpha1).t()
        mrd_Dphasing = (mrd_Dphasing.t() / eta).t()

        return mrd_phasing, mrd_Dphasing

    def phenom_d_int_phase(self, Mf, eta, eta2, xi):
        beta1 = self.beta1Fit(eta, eta2, xi)
        beta2 = self.beta2Fit(eta, eta2, xi)
        beta3 = self.beta3Fit(eta, eta2, xi)
        # Merger phase
        # Leading beta0 is not added here
        # overall 1/eta is not multiplied
        int_phasing = (Mf.mT * beta1).mT
        int_phasing += (torch.log(Mf).mT * beta2).mT
        int_phasing -= (Mf.mT ** (-3.0) / 3.0 * beta3).mT

        # overall 1/eta is multiple in derivative of
        # intermediate phase
        int_Dphasing = (Mf.mT ** (-4.0) * beta3).mT
        int_Dphasing += (Mf.mT ** (-1.0) * beta2).mT
        int_Dphasing = (int_Dphasing.mT + beta1).mT
        int_Dphasing = (int_Dphasing.mT / eta).mT
        return int_phasing, int_Dphasing

    def subtract3PNSS(self, Mf, mass1, mass2, eta, eta2, xi, chi1, chi2):
        M = mass1 + mass2
        eta = mass1 * mass2 / M / M
        m1byM = mass1 / M
        m2byM = mass2 / M
        chi1sq = chi1 * chi1
        chi2sq = chi2 * chi2
        v1 = (PI * Mf) ** (1.0 / 3.0)
        v5 = v1**5.0
        v6 = v1**6.0
        v7 = v1**7.0
        pfaN = 3.0 / (128.0 * eta)
        pn_ss3 = (
            (326.75 / 1.12 + 557.5 / 1.8 * eta) * eta * chi1 * chi2
            + (
                (4703.5 / 8.4 + 2935.0 / 6.0 * m1byM - 120.0 * m1byM * m1byM)
                * m1byM
                * m1byM
                + (
                    -4108.25 / 6.72
                    - 108.5 / 1.2 * m1byM
                    + 125.5 / 3.6 * m1byM * m1byM
                )
                * m1byM
                * m1byM
            )
            * chi1sq
            + (
                (4703.5 / 8.4 + 2935.0 / 6.0 * m2byM - 120.0 * m2byM * m2byM)
                * m2byM
                * m2byM
                + (
                    -4108.25 / 6.72
                    - 108.5 / 1.2 * m2byM
                    + 125.5 / 3.6 * m2byM * m2byM
                )
                * m2byM
                * m2byM
            )
            * chi2sq
        )
        phase_pn_ss3 = (((v6.mT * pn_ss3).mT / v5).mT * pfaN).mT
        Dphase_pn_ss3 = (
            (PI * (v6.mT * pn_ss3).mT / (3.0 * v1 * v7)).mT * pfaN
        ).mT
        return phase_pn_ss3, Dphase_pn_ss3

    def phenom_d_inspiral_phase(
        self, Mf, mass_1, mass_2, eta, eta2, xi, chi1, chi2
    ):
        ins_phasing, ins_Dphasing = self.taylorf2_phase(
            Mf, mass_1, mass_2, chi1, chi2
        )
        # subtract 3PN spin-spin term as this is in LAL's TaylorF2
        # implementation, but was not available when PhenomD was tuned.
        # refer https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimIMRPhenomD.c#L397-398 # noqa: E501
        pn_ss3, Dpn_ss3 = self.subtract3PNSS(
            Mf, mass_1, mass_2, eta, eta2, xi, chi1, chi2
        )
        ins_phasing -= pn_ss3
        ins_Dphasing -= Dpn_ss3

        sigma1 = self.sigma1Fit(eta, eta2, xi)
        sigma2 = self.sigma2Fit(eta, eta2, xi)
        sigma3 = self.sigma3Fit(eta, eta2, xi)
        sigma4 = self.sigma4Fit(eta, eta2, xi)

        ins_phasing += (Mf.mT * sigma1 / eta).mT
        ins_phasing += (Mf.mT ** (4.0 / 3.0) * 0.75 * sigma2 / eta).mT
        ins_phasing += (Mf.mT ** (5.0 / 3.0) * 0.6 * sigma3 / eta).mT
        ins_phasing += (Mf.mT**2.0 * 0.5 * sigma4 / eta).mT

        ins_Dphasing = (ins_Dphasing.mT + sigma1 / eta).mT
        ins_Dphasing += (Mf.mT ** (1.0 / 3.0) * sigma2 / eta).mT
        ins_Dphasing += (Mf.mT ** (2.0 / 3.0) * sigma3 / eta).mT
        ins_Dphasing += (Mf.mT * sigma4 / eta).mT

        return ins_phasing, ins_Dphasing

    def fring_fdamp(self, eta, eta2, chi1, chi2):
        finspin = self.FinalSpin0815(eta, eta2, chi1, chi2)
        Erad = self.PhenomInternal_EradRational0815(eta, eta2, chi1, chi2)

        fRD, fDM = self._linear_interp_finspin(finspin)
        fRD /= 1.0 - Erad
        fDM /= 1.0 - Erad

        return fRD, fDM

    def fmaxCalc(self, fRD, fDM, gamma2, gamma3):
        mask = gamma2 <= 1
        # calculate result for gamma2 <= 1 case
        sqrt_term = torch.sqrt(1 - gamma2.pow(2))
        result_case1 = fRD + (fDM * (-1 + sqrt_term) * gamma3) / gamma2

        # calculate result for gamma2 > 1 case
        # i.e. don't add sqrt term
        result_case2 = fRD + (-fDM * gamma3) / gamma2

        # combine results using mask
        result = torch.where(mask, result_case1, result_case2)

        return torch.abs(result)

    def _linear_interp_finspin(self, finspin):
        # chi is a batch of final spins i.e. torch.Size([n])
        right_spin_idx = torch.bucketize(finspin, self.qnmdata_a)
        right_spin_val = self.qnmdata_a[right_spin_idx]
        # QNMData_a is sorted, hence take the previous index
        left_spin_idx = right_spin_idx - 1
        left_spin_val = self.qnmdata_a[left_spin_idx]

        if not torch.all(left_spin_val < right_spin_val):
            raise RuntimeError(
                "Left value in grid should be greater than right. "
                "Maybe be caused for extremal spin values."
            )
        left_fring = self.qnmdata_fring[left_spin_idx]
        right_fring = self.qnmdata_fring[right_spin_idx]
        slope_fring = right_fring - left_fring
        slope_fring /= right_spin_val - left_spin_val

        left_fdamp = self.qnmdata_fdamp[left_spin_idx]
        right_fdamp = self.qnmdata_fdamp[right_spin_idx]
        slope_fdamp = right_fdamp - left_fdamp
        slope_fdamp /= right_spin_val - left_spin_val

        return (
            slope_fring * (finspin - left_spin_val) + left_fring,
            slope_fdamp * (finspin - left_spin_val) + left_fdamp,
        )

    # Utility functions taken from PhenomD utilities in lalsimulation
    # https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimIMRPhenomD_internals.c
    def sigma1Fit(self, eta, eta2, xi):
        return (
            2096.551999295543
            + 1463.7493168261553 * eta
            + (
                1312.5493286098522
                + 18307.330017082117 * eta
                - 43534.1440746107 * eta2
                + (
                    -833.2889543511114
                    + 32047.31997183187 * eta
                    - 108609.45037520859 * eta2
                )
                * xi
                + (
                    452.25136398112204
                    + 8353.439546391714 * eta
                    - 44531.3250037322 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def sigma2Fit(self, eta, eta2, xi):
        return (
            -10114.056472621156
            - 44631.01109458185 * eta
            + (
                -6541.308761668722
                - 266959.23419307504 * eta
                + 686328.3229317984 * eta2
                + (
                    3405.6372187679685
                    - 437507.7208209015 * eta
                    + 1.6318171307344697e6 * eta2
                )
                * xi
                + (
                    -7462.648563007646
                    - 114585.25177153319 * eta
                    + 674402.4689098676 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def sigma3Fit(self, eta, eta2, xi):
        return (
            22933.658273436497
            + 230960.00814979506 * eta
            + (
                14961.083974183695
                + 1.1940181342318142e6 * eta
                - 3.1042239693052764e6 * eta2
                + (
                    -3038.166617199259
                    + 1.8720322849093592e6 * eta
                    - 7.309145012085539e6 * eta2
                )
                * xi
                + (
                    42738.22871475411
                    + 467502.018616601 * eta
                    - 3.064853498512499e6 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def sigma4Fit(self, eta, eta2, xi):
        return (
            -14621.71522218357
            - 377812.8579387104 * eta
            + (
                -9608.682631509726
                - 1.7108925257214056e6 * eta
                + 4.332924601416521e6 * eta2
                + (
                    -22366.683262266528
                    - 2.5019716386377467e6 * eta
                    + 1.0274495902259542e7 * eta2
                )
                * xi
                + (
                    -85360.30079034246
                    - 570025.3441737515 * eta
                    + 4.396844346849777e6 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def gamma1_fun(self, eta, eta2, xi):
        return (
            0.006927402739328343
            + 0.03020474290328911 * eta
            + (
                0.006308024337706171
                - 0.12074130661131138 * eta
                + 0.26271598905781324 * eta2
                + (
                    0.0034151773647198794
                    - 0.10779338611188374 * eta
                    + 0.27098966966891747 * eta2
                )
                * xi
                + (
                    0.0007374185938559283
                    - 0.02749621038376281 * eta
                    + 0.0733150789135702 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def gamma2_fun(self, eta, eta2, xi):
        return (
            1.010344404799477
            + 0.0008993122007234548 * eta
            + (
                0.283949116804459
                - 4.049752962958005 * eta
                + 13.207828172665366 * eta2
                + (
                    0.10396278486805426
                    - 7.025059158961947 * eta
                    + 24.784892370130475 * eta2
                )
                * xi
                + (
                    0.03093202475605892
                    - 2.6924023896851663 * eta
                    + 9.609374464684983 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def gamma3_fun(self, eta, eta2, xi):
        return (
            1.3081615607036106
            - 0.005537729694807678 * eta
            + (
                -0.06782917938621007
                - 0.6689834970767117 * eta
                + 3.403147966134083 * eta2
                + (
                    -0.05296577374411866
                    - 0.9923793203111362 * eta
                    + 4.820681208409587 * eta2
                )
                * xi
                + (
                    -0.006134139870393713
                    - 0.38429253308696365 * eta
                    + 1.7561754421985984 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def beta1Fit(self, eta, eta2, xi):
        return (
            97.89747327985583
            - 42.659730877489224 * eta
            + (
                153.48421037904913
                - 1417.0620760768954 * eta
                + 2752.8614143665027 * eta2
                + (
                    138.7406469558649
                    - 1433.6585075135881 * eta
                    + 2857.7418952430758 * eta2
                )
                * xi
                + (
                    41.025109467376126
                    - 423.680737974639 * eta
                    + 850.3594335657173 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def beta2Fit(self, eta, eta2, xi):
        return (
            -3.282701958759534
            - 9.051384468245866 * eta
            + (
                -12.415449742258042
                + 55.4716447709787 * eta
                - 106.05109938966335 * eta2
                + (
                    -11.953044553690658
                    + 76.80704618365418 * eta
                    - 155.33172948098394 * eta2
                )
                * xi
                + (
                    -3.4129261592393263
                    + 25.572377569952536 * eta
                    - 54.408036707740465 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def beta3Fit(self, eta, eta2, xi):
        return (
            -0.000025156429818799565
            + 0.000019750256942201327 * eta
            + (
                -0.000018370671469295915
                + 0.000021886317041311973 * eta
                + 0.00008250240316860033 * eta2
                + (
                    7.157371250566708e-6
                    - 0.000055780000112270685 * eta
                    + 0.00019142082884072178 * eta2
                )
                * xi
                + (
                    5.447166261464217e-6
                    - 0.00003220610095021982 * eta
                    + 0.00007974016714984341 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def alpha1Fit(self, eta, eta2, xi):
        return (
            43.31514709695348
            + 638.6332679188081 * eta
            + (
                -32.85768747216059
                + 2415.8938269370315 * eta
                - 5766.875169379177 * eta2
                + (
                    -61.85459307173841
                    + 2953.967762459948 * eta
                    - 8986.29057591497 * eta2
                )
                * xi
                + (
                    -21.571435779762044
                    + 981.2158224673428 * eta
                    - 3239.5664895930286 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def alpha2Fit(self, eta, eta2, xi):
        return (
            -0.07020209449091723
            - 0.16269798450687084 * eta
            + (
                -0.1872514685185499
                + 1.138313650449945 * eta
                - 2.8334196304430046 * eta2
                + (
                    -0.17137955686840617
                    + 1.7197549338119527 * eta
                    - 4.539717148261272 * eta2
                )
                * xi
                + (
                    -0.049983437357548705
                    + 0.6062072055948309 * eta
                    - 1.682769616644546 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def alpha3Fit(self, eta, eta2, xi):
        return (
            9.5988072383479
            - 397.05438595557433 * eta
            + (
                16.202126189517813
                - 1574.8286986717037 * eta
                + 3600.3410843831093 * eta2
                + (
                    27.092429659075467
                    - 1786.482357315139 * eta
                    + 5152.919378666511 * eta2
                )
                * xi
                + (
                    11.175710130033895
                    - 577.7999423177481 * eta
                    + 1808.730762932043 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def alpha4Fit(self, eta, eta2, xi):
        return (
            -0.02989487384493607
            + 1.4022106448583738 * eta
            + (
                -0.07356049468633846
                + 0.8337006542278661 * eta
                + 0.2240008282397391 * eta2
                + (
                    -0.055202870001177226
                    + 0.5667186343606578 * eta
                    + 0.7186931973380503 * eta2
                )
                * xi
                + (
                    -0.015507437354325743
                    + 0.15750322779277187 * eta
                    + 0.21076815715176228 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def alpha5Fit(self, eta, eta2, xi):
        return (
            0.9974408278363099
            - 0.007884449714907203 * eta
            + (
                -0.059046901195591035
                + 1.3958712396764088 * eta
                - 4.516631601676276 * eta2
                + (
                    -0.05585343136869692
                    + 1.7516580039343603 * eta
                    - 5.990208965347804 * eta2
                )
                * xi
                + (
                    -0.017945336522161195
                    + 0.5965097794825992 * eta
                    - 2.0608879367971804 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def rho1_fun(self, eta, eta2, xi):
        return (
            3931.8979897196696
            - 17395.758706812805 * eta
            + (
                3132.375545898835
                + 343965.86092361377 * eta
                - 1.2162565819981997e6 * eta2
                + (
                    -70698.00600428853
                    + 1.383907177859705e6 * eta
                    - 3.9662761890979446e6 * eta2
                )
                * xi
                + (
                    -60017.52423652596
                    + 803515.1181825735 * eta
                    - 2.091710365941658e6 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def rho2_fun(self, eta, eta2, xi):
        return (
            -40105.47653771657
            + 112253.0169706701 * eta
            + (
                23561.696065836168
                - 3.476180699403351e6 * eta
                + 1.137593670849482e7 * eta2
                + (
                    754313.1127166454
                    - 1.308476044625268e7 * eta
                    + 3.6444584853928134e7 * eta2
                )
                * xi
                + (
                    596226.612472288
                    - 7.4277901143564405e6 * eta
                    + 1.8928977514040343e7 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def rho3_fun(self, eta, eta2, xi):
        return (
            83208.35471266537
            - 191237.7264145924 * eta
            + (
                -210916.2454782992
                + 8.71797508352568e6 * eta
                - 2.6914942420669552e7 * eta2
                + (
                    -1.9889806527362722e6
                    + 3.0888029960154563e7 * eta
                    - 8.390870279256162e7 * eta2
                )
                * xi
                + (
                    -1.4535031953446497e6
                    + 1.7063528990822166e7 * eta
                    - 4.2748659731120914e7 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def FinalSpin0815(self, eta, eta2, chi1, chi2):
        Seta = torch.sqrt(1.0 - 4.0 * eta)
        Seta = torch.nan_to_num(Seta)  # avoid nan around eta = 0.25
        m1 = 0.5 * (1.0 + Seta)
        m2 = 0.5 * (1.0 - Seta)
        m1s = m1 * m1
        m2s = m2 * m2
        s = m1s * chi1 + m2s * chi2
        eta3 = eta2 * eta
        s2 = s * s
        s3 = s2 * s
        return eta * (
            3.4641016151377544
            - 4.399247300629289 * eta
            + 9.397292189321194 * eta2
            - 13.180949901606242 * eta3
            + (
                (1.0 / eta - 0.0850917821418767 - 5.837029316602263 * eta)
                + (0.1014665242971878 - 2.0967746996832157 * eta) * s
                + (-1.3546806617824356 + 4.108962025369336 * eta) * s2
                + (-0.8676969352555539 + 2.064046835273906 * eta) * s3
            )
            * s
        )

    def PhenomInternal_EradRational0815(self, eta, eta2, chi1, chi2):
        Seta = torch.sqrt(1.0 - 4.0 * eta)
        m1 = 0.5 * (1.0 + Seta)
        m2 = 0.5 * (1.0 - Seta)
        m1s = m1 * m1
        m2s = m2 * m2
        s = (m1s * chi1 + m2s * chi2) / (m1s + m2s)

        eta3 = eta2 * eta

        return (
            eta
            * (
                0.055974469826360077
                + 0.5809510763115132 * eta
                - 0.9606726679372312 * eta2
                + 3.352411249771192 * eta3
            )
            * (
                1.0
                + (
                    -0.0030302335878845507
                    - 2.0066110851351073 * eta
                    + 7.7050567802399215 * eta2
                )
                * s
            )
        ) / (
            1.0
            + (
                -0.6714403054720589
                - 1.4756929437702908 * eta
                + 7.304676214885011 * eta2
            )
            * s
        )

    def AmpIntColFitCoeff(self, eta, eta2, xi):
        return (
            0.8149838730507785
            + 2.5747553517454658 * eta
            + (
                1.1610198035496786
                - 2.3627771785551537 * eta
                + 6.771038707057573 * eta2
                + (
                    0.7570782938606834
                    - 2.7256896890432474 * eta
                    + 7.1140380397149965 * eta2
                )
                * xi
                + (
                    0.1766934149293479
                    - 0.7978690983168183 * eta
                    + 2.1162391502005153 * eta2
                )
                * xi
                * xi
            )
            * xi
        )

    def delta_values(self, f1, f2, f3, v1, v2, v3, d1, d2):
        f12 = f1 * f1
        f13 = f1 * f12
        f14 = f1 * f13
        f15 = f1 * f14
        f22 = f2 * f2
        f23 = f2 * f22
        f24 = f2 * f23
        f32 = f3 * f3
        f33 = f3 * f32
        f34 = f3 * f33
        f35 = f3 * f34
        delta_0 = -(
            (
                d2 * f15 * f22 * f3
                - 2 * d2 * f14 * f23 * f3
                + d2 * f13 * f24 * f3
                - d2 * f15 * f2 * f32
                + d2 * f14 * f22 * f32
                - d1 * f13 * f23 * f32
                + d2 * f13 * f23 * f32
                + d1 * f12 * f24 * f32
                - d2 * f12 * f24 * f32
                + d2 * f14 * f2 * f33
                + 2 * d1 * f13 * f22 * f33
                - 2 * d2 * f13 * f22 * f33
                - d1 * f12 * f23 * f33
                + d2 * f12 * f23 * f33
                - d1 * f1 * f24 * f33
                - d1 * f13 * f2 * f34
                - d1 * f12 * f22 * f34
                + 2 * d1 * f1 * f23 * f34
                + d1 * f12 * f2 * f35
                - d1 * f1 * f22 * f35
                + 4 * f12 * f23 * f32 * v1
                - 3 * f1 * f24 * f32 * v1
                - 8 * f12 * f22 * f33 * v1
                + 4 * f1 * f23 * f33 * v1
                + f24 * f33 * v1
                + 4 * f12 * f2 * f34 * v1
                + f1 * f22 * f34 * v1
                - 2 * f23 * f34 * v1
                - 2 * f1 * f2 * f35 * v1
                + f22 * f35 * v1
                - f15 * f32 * v2
                + 3 * f14 * f33 * v2
                - 3 * f13 * f34 * v2
                + f12 * f35 * v2
                - f15 * f22 * v3
                + 2 * f14 * f23 * v3
                - f13 * f24 * v3
                + 2 * f15 * f2 * f3 * v3
                - f14 * f22 * f3 * v3
                - 4 * f13 * f23 * f3 * v3
                + 3 * f12 * f24 * f3 * v3
                - 4 * f14 * f2 * f32 * v3
                + 8 * f13 * f22 * f32 * v3
                - 4 * f12 * f23 * f32 * v3
            )
            / ((f1 - f2) ** 2 * (f1 - f3) ** 3 * (f3 - f2) ** 2)
        )

        delta_1 = -(
            (
                -(d2 * f15 * f22)
                + 2 * d2 * f14 * f23
                - d2 * f13 * f24
                - d2 * f14 * f22 * f3
                + 2 * d1 * f13 * f23 * f3
                + 2 * d2 * f13 * f23 * f3
                - 2 * d1 * f12 * f24 * f3
                - d2 * f12 * f24 * f3
                + d2 * f15 * f32
                - 3 * d1 * f13 * f22 * f32
                - d2 * f13 * f22 * f32
                + 2 * d1 * f12 * f23 * f32
                - 2 * d2 * f12 * f23 * f32
                + d1 * f1 * f24 * f32
                + 2 * d2 * f1 * f24 * f32
                - d2 * f14 * f33
                + d1 * f12 * f22 * f33
                + 3 * d2 * f12 * f22 * f33
                - 2 * d1 * f1 * f23 * f33
                - 2 * d2 * f1 * f23 * f33
                + d1 * f24 * f33
                + d1 * f13 * f34
                + d1 * f1 * f22 * f34
                - 2 * d1 * f23 * f34
                - d1 * f12 * f35
                + d1 * f22 * f35
                - 8 * f12 * f23 * f3 * v1
                + 6 * f1 * f24 * f3 * v1
                + 12 * f12 * f22 * f32 * v1
                - 8 * f1 * f23 * f32 * v1
                - 4 * f12 * f34 * v1
                + 2 * f1 * f35 * v1
                + 2 * f15 * f3 * v2
                - 4 * f14 * f32 * v2
                + 4 * f12 * f34 * v2
                - 2 * f1 * f35 * v2
                - 2 * f15 * f3 * v3
                + 8 * f12 * f23 * f3 * v3
                - 6 * f1 * f24 * f3 * v3
                + 4 * f14 * f32 * v3
                - 12 * f12 * f22 * f32 * v3
                + 8 * f1 * f23 * f32 * v3
            )
            / ((f1 - f2) ** 2 * (f1 - f3) ** 3 * (-f2 + f3) ** 2)
        )

        delta_2 = -(
            (
                d2 * f15 * f2
                - d1 * f13 * f23
                - 3 * d2 * f13 * f23
                + d1 * f12 * f24
                + 2 * d2 * f12 * f24
                - d2 * f15 * f3
                + d2 * f14 * f2 * f3
                - d1 * f12 * f23 * f3
                + d2 * f12 * f23 * f3
                + d1 * f1 * f24 * f3
                - d2 * f1 * f24 * f3
                - d2 * f14 * f32
                + 3 * d1 * f13 * f2 * f32
                + d2 * f13 * f2 * f32
                - d1 * f1 * f23 * f32
                + d2 * f1 * f23 * f32
                - 2 * d1 * f24 * f32
                - d2 * f24 * f32
                - 2 * d1 * f13 * f33
                + 2 * d2 * f13 * f33
                - d1 * f12 * f2 * f33
                - 3 * d2 * f12 * f2 * f33
                + 3 * d1 * f23 * f33
                + d2 * f23 * f33
                + d1 * f12 * f34
                - d1 * f1 * f2 * f34
                + d1 * f1 * f35
                - d1 * f2 * f35
                + 4 * f12 * f23 * v1
                - 3 * f1 * f24 * v1
                + 4 * f1 * f23 * f3 * v1
                - 3 * f24 * f3 * v1
                - 12 * f12 * f2 * f32 * v1
                + 4 * f23 * f32 * v1
                + 8 * f12 * f33 * v1
                - f1 * f34 * v1
                - f35 * v1
                - f15 * v2
                - f14 * f3 * v2
                + 8 * f13 * f32 * v2
                - 8 * f12 * f33 * v2
                + f1 * f34 * v2
                + f35 * v2
                + f15 * v3
                - 4 * f12 * f23 * v3
                + 3 * f1 * f24 * v3
                + f14 * f3 * v3
                - 4 * f1 * f23 * f3 * v3
                + 3 * f24 * f3 * v3
                - 8 * f13 * f32 * v3
                + 12 * f12 * f2 * f32 * v3
                - 4 * f23 * f32 * v3
            )
            / ((f1 - f2) ** 2 * (f1 - f3) ** 3 * (-f2 + f3) ** 2)
        )

        delta_3 = -(
            (
                -2 * d2 * f14 * f2
                + d1 * f13 * f22
                + 3 * d2 * f13 * f22
                - d1 * f1 * f24
                - d2 * f1 * f24
                + 2 * d2 * f14 * f3
                - 2 * d1 * f13 * f2 * f3
                - 2 * d2 * f13 * f2 * f3
                + d1 * f12 * f22 * f3
                - d2 * f12 * f22 * f3
                + d1 * f24 * f3
                + d2 * f24 * f3
                + d1 * f13 * f32
                - d2 * f13 * f32
                - 2 * d1 * f12 * f2 * f32
                + 2 * d2 * f12 * f2 * f32
                + d1 * f1 * f22 * f32
                - d2 * f1 * f22 * f32
                + d1 * f12 * f33
                - d2 * f12 * f33
                + 2 * d1 * f1 * f2 * f33
                + 2 * d2 * f1 * f2 * f33
                - 3 * d1 * f22 * f33
                - d2 * f22 * f33
                - 2 * d1 * f1 * f34
                + 2 * d1 * f2 * f34
                - 4 * f12 * f22 * v1
                + 2 * f24 * v1
                + 8 * f12 * f2 * f3 * v1
                - 4 * f1 * f22 * f3 * v1
                - 4 * f12 * f32 * v1
                + 8 * f1 * f2 * f32 * v1
                - 4 * f22 * f32 * v1
                - 4 * f1 * f33 * v1
                + 2 * f34 * v1
                + 2 * f14 * v2
                - 4 * f13 * f3 * v2
                + 4 * f1 * f33 * v2
                - 2 * f34 * v2
                - 2 * f14 * v3
                + 4 * f12 * f22 * v3
                - 2 * f24 * v3
                + 4 * f13 * f3 * v3
                - 8 * f12 * f2 * f3 * v3
                + 4 * f1 * f22 * f3 * v3
                + 4 * f12 * f32 * v3
                - 8 * f1 * f2 * f32 * v3
                + 4 * f22 * f32 * v3
            )
            / ((f1 - f2) ** 2 * (f1 - f3) ** 3 * (-f2 + f3) ** 2)
        )

        delta_4 = -(
            (
                d2 * f13 * f2
                - d1 * f12 * f22
                - 2 * d2 * f12 * f22
                + d1 * f1 * f23
                + d2 * f1 * f23
                - d2 * f13 * f3
                + 2 * d1 * f12 * f2 * f3
                + d2 * f12 * f2 * f3
                - d1 * f1 * f22 * f3
                + d2 * f1 * f22 * f3
                - d1 * f23 * f3
                - d2 * f23 * f3
                - d1 * f12 * f32
                + d2 * f12 * f32
                - d1 * f1 * f2 * f32
                - 2 * d2 * f1 * f2 * f32
                + 2 * d1 * f22 * f32
                + d2 * f22 * f32
                + d1 * f1 * f33
                - d1 * f2 * f33
                + 3 * f1 * f22 * v1
                - 2 * f23 * v1
                - 6 * f1 * f2 * f3 * v1
                + 3 * f22 * f3 * v1
                + 3 * f1 * f32 * v1
                - f33 * v1
                - f13 * v2
                + 3 * f12 * f3 * v2
                - 3 * f1 * f32 * v2
                + f33 * v2
                + f13 * v3
                - 3 * f1 * f22 * v3
                + 2 * f23 * v3
                - 3 * f12 * f3 * v3
                + 6 * f1 * f2 * f3 * v3
                - 3 * f22 * f3 * v3
            )
            / ((f1 - f2) ** 2 * (f1 - f3) ** 3 * (-f2 + f3) ** 2)
        )

        return delta_0, delta_1, delta_2, delta_3, delta_4

    def chiPN(self, Seta, eta, chi1, chi2):
        chi_s = chi1 + chi2
        chi_a = chi1 - chi2

        return 0.5 * (chi_s * (1.0 - eta * 76.0 / 113.0) + Seta * chi_a)



from collections.abc import Iterable

import torch

from ..types import WaveformTensor


class ChunkedTimeSeriesDataset(torch.utils.data.IterableDataset):
    """
    Wrapper dataset that will loop through chunks of timeseries
    data produced by another iterable and sample windows from
    these chunks.

    Args:
        chunk_it:
            Iterator which will produce chunks of timeseries
            data to sample windows from. Should have shape
            `(N, C, T)`, where `N` is the number of chunks
            to sample from, `C` is the number of channels,
            and `T` is the number of samples along the
            time dimension for each chunk.
        kernel_size:
            Size of windows to be sampled from each chunk.
            Should be less than the size of each chunk
            along the time dimension.
        batch_size:
            Number of windows to sample at each iteration
        batches_per_chunk:
            Number of batches of windows to sample from
            each chunk before moving on to the next one.
            Sampling fewer batches from each chunk means
            a lower likelihood of sampling duplicate windows,
            but an increase in chunk-loading overhead.
        coincident:
            Whether the windows sampled from individual
            channels in each batch element should be
            sampled coincidentally, i.e. consisting of
            the same timesteps, or whether each window
            should be sample independently from the others.
        device:
            Which device chunks should be moved to upon loading.
    """

    def __init__(
        self,
        chunk_it: Iterable,
        kernel_size: float,
        batch_size: int,
        batches_per_chunk: int,
        coincident: bool = True,
        device: str = "cpu",
    ) -> None:
        self.chunk_it = chunk_it
        self.kernel_size = kernel_size
        self.batch_size = batch_size
        self.batches_per_chunk = batches_per_chunk
        self.coincident = coincident
        self.device = device

    def __len__(self) -> int:
        return len(self.chunk_it) * self.batches_per_chunk

    def __iter__(self) -> WaveformTensor:
        it = iter(self.chunk_it)
        chunk = next(it)
        num_chunks, num_channels, chunk_size = chunk.shape

        # if we're sampling coincidentally, we only need
        # to sample indices on a per-batch-element basis.
        # Otherwise, we'll need indices for both each
        # batch sample _and_ each channel with each sample
        if self.coincident:
            sample_size = (self.batch_size,)
        else:
            sample_size = (self.batch_size, num_channels)

        # slice kernels out a flattened chunk tensor
        # index-for-index. We'll account for batch/
        # channel indices by introducing offsets later on
        idx = torch.arange(self.kernel_size, device=self.device)
        idx = idx.view(1, 1, -1)
        idx = idx.repeat(self.batch_size, num_channels, 1)

        # this will just be a set of aranged channel indices
        # repeated to offset the kernel indices in the
        # flattened chunk tensor
        channel_idx = torch.arange(num_channels, device=self.device)
        channel_idx = channel_idx.view(1, -1, 1)
        channel_idx = channel_idx.repeat(self.batch_size, 1, self.kernel_size)
        idx += channel_idx * chunk_size

        while True:
            # record the number of rows in the chunk, then
            # flatten it to make it easier to slice
            if chunk_size < self.kernel_size:
                raise ValueError(
                    "Can't sample kernels of size {} from chunk "
                    "with size {}".format(self.kernel_size, chunk_size)
                )
            chunk = chunk.reshape(-1)

            # generate batches from the current chunk
            for _ in range(self.batches_per_chunk):
                # first sample the indices of which chunk elements
                # we're going to read batch elements from
                chunk_idx = torch.randint(
                    0, num_chunks, size=sample_size, device=self.device
                )

                # account for the offset this batch element
                # introduces in the flattened array
                chunk_idx *= num_channels * chunk_size
                chunk_idx = chunk_idx.view(self.batch_size, -1, 1)
                chunk_idx = chunk_idx + idx

                # now sample the start index within each chunk
                # element we're going to grab our time windows from
                time_idx = torch.randint(
                    0,
                    chunk_size - self.kernel_size,
                    size=sample_size,
                    device=self.device,
                )
                time_idx = time_idx.view(self.batch_size, -1, 1)

                # there's no additional offset factor to account for here
                chunk_idx += time_idx

                # now slice this 3D tensor from our flattened chunk
                yield chunk[chunk_idx]

            try:
                chunk = next(it)
            except StopIteration:
                break
            num_chunks, num_channels, chunk_size = chunk.shape



from typing import Union

from jaxtyping import Float
from torch import Tensor

WaveformTensor = Float[Tensor, "batch num_ifos time"]
PSDTensor = Float[Tensor, "num_ifos frequency"]
BatchTensor = Float[Tensor, "batch"]
VectorGeometry = Float[Tensor, "batch space"]
TensorGeometry = Float[Tensor, "batch space space"]
NetworkVertices = Float[Tensor, "num_ifos 3"]
NetworkDetectorTensors = Float[Tensor, "num_ifos 3 3"]


TimeSeries1d = Float[Tensor, "time"]
TimeSeries2d = Float[TimeSeries1d, "channel"]
TimeSeries3d = Float[TimeSeries2d, "batch"]
TimeSeries1to3d = Union[TimeSeries1d, TimeSeries2d, TimeSeries3d]

FrequencySeries1d = Float[Tensor, "frequency"]
FrequencySeries2d = Float[FrequencySeries1d, "channel"]
FrequencySeries3d = Float[FrequencySeries2d, "batch"]
FrequencySeries1to3d = Union[
    FrequencySeries1d, FrequencySeries2d, FrequencySeries3d
]



import torch
from torch import Tensor

from ml4gw.types import BatchTensor


def semi_major_minor_from_e(e: Tensor):
    a = 1.0 / torch.sqrt(2.0 - (e * e))
    b = a * torch.sqrt(1.0 - (e * e))
    return a, b


class SineGaussian(torch.nn.Module):
    """
    Callable class for generating sine-Gaussian waveforms.

    Args:
        sample_rate: Sample rate of waveform
        duration: Duration of waveform
    """

    def __init__(self, sample_rate: float, duration: float):
        super().__init__()
        # determine times based on requested duration and sample rate
        # and shift so that the waveform is centered at t=0

        num = int(duration * sample_rate)
        times = torch.arange(num, dtype=torch.float64) / sample_rate
        times -= duration / 2.0

        self.register_buffer("times", times)

    def forward(
        self,
        quality: BatchTensor,
        frequency: BatchTensor,
        hrss: BatchTensor,
        phase: BatchTensor,
        eccentricity: BatchTensor,
    ):
        """
        Generate lalinference implementation of a sine-Gaussian waveform.
        See
        git.ligo.org/lscsoft/lalsuite/-/blob/master/lalinference/lib/LALInferenceBurstRoutines.c#L381
        for details on parameter definitions.

        Args:
            frequency:
                Central frequency of the sine-Gaussian waveform
            quality:
                Quality factor of the sine-Gaussian waveform
            hrss:
                Hrss of the sine-Gaussian waveform
            phase:
                Phase of the sine-Gaussian waveform
            eccentricity:
                Eccentricity of the sine-Gaussian waveform.
                Controls the relative amplitudes of the
                hplus and hcross polarizations.
        Returns:
            Tensors of cross and plus polarizations
        """
        dtype = frequency.dtype
        # add dimension for calculating waveforms in batch
        frequency = frequency.view(-1, 1)
        quality = quality.view(-1, 1)
        hrss = hrss.view(-1, 1)
        phase = phase.view(-1, 1)
        eccentricity = eccentricity.view(-1, 1)

        # TODO: enforce all inputs are on the same device?
        pi = torch.tensor([torch.pi], device=frequency.device)

        # calculate relative hplus / hcross amplitudes based on eccentricity
        # as well as normalization factors
        a, b = semi_major_minor_from_e(eccentricity)
        norm_prefactor = quality / (4.0 * frequency * torch.sqrt(pi))
        cosine_norm = norm_prefactor * (1.0 + torch.exp(-quality * quality))
        sine_norm = norm_prefactor * (1.0 - torch.exp(-quality * quality))

        cos_phase, sin_phase = torch.cos(phase), torch.sin(phase)

        h0_plus = (
            hrss
            * a
            / torch.sqrt(
                cosine_norm * (cos_phase**2) + sine_norm * (sin_phase**2)
            )
        )
        h0_cross = (
            hrss
            * b
            / torch.sqrt(
                cosine_norm * (sin_phase**2) + sine_norm * (cos_phase**2)
            )
        )

        # cast the phase to a complex number
        phi = 2 * pi * frequency * self.times
        complex_phase = torch.complex(torch.zeros_like(phi), (phi - phase))

        # calculate the waveform and apply a tukey window to taper the waveform
        fac = torch.exp(phi**2 / (-2.0 * quality**2) + complex_phase)

        cross = fac.imag * h0_cross
        plus = fac.real * h0_plus

        cross = cross.to(dtype)
        plus = plus.to(dtype)

        return cross, plus
